"""
NOTE: The color names to hex are copied from the color sheets from manim

MIT License

Copyright (c) 2018 3Blue1Brown LLC
Copyright (c) 2021, the Manim Community Developers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

colors = {
    'WHITE': '#FFFFFF',
    'GRAY_A': '#DDDDDD',
    'GREY_A': '#DDDDDD',
    'GRAY_B': '#BBBBBB',
    'GREY_B': '#BBBBBB',
    'GRAY_C': '#888888',
    'GREY_C': '#888888',
    'GRAY_D': '#444444',
    'GREY_D': '#444444',
    'GRAY_E': '#222222',
    'GREY_E': '#222222',
    'BLACK': '#000000',
    'LIGHTER_GRAY': '#DDDDDD',
    'LIGHTER_GREY': '#DDDDDD',
    'LIGHT_GRAY': '#BBBBBB',
    'LIGHT_GREY': '#BBBBBB',
    'GRAY': '#888888',
    'GREY': '#888888',
    'DARK_GRAY': '#444444',
    'DARK_GREY': '#444444',
    'DARKER_GRAY': '#222222',
    'DARKER_GREY': '#222222',
    'BLUE_A': '#C7E9F1',
    'BLUE_B': '#9CDCEB',
    'BLUE_C': '#58C4DD',
    'BLUE_D': '#29ABCA',
    'BLUE_E': '#236B8E',
    'PURE_BLUE': '#0000FF',
    'BLUE': '#58C4DD',
    'DARK_BLUE': '#236B8E',
    'TEAL_A': '#ACEAD7',
    'TEAL_B': '#76DDC0',
    'TEAL_C': '#5CD0B3',
    'TEAL_D': '#55C1A7',
    'TEAL_E': '#49A88F',
    'TEAL': '#5CD0B3',
    'GREEN_A': '#C9E2AE',
    'GREEN_B': '#A6CF8C',
    'GREEN_C': '#83C167',
    'GREEN_D': '#77B05D',
    'GREEN_E': '#699C52',
    'PURE_GREEN': '#00FF00',
    'GREEN': '#83C167',
    'YELLOW_A': '#FFF1B6',
    'YELLOW_B': '#FFEA94',
    'YELLOW_C': '#FFFF00',
    'YELLOW_D': '#F4D345',
    'YELLOW_E': '#E8C11C',
    'YELLOW': '#FFFF00',
    'GOLD_A': '#F7C797',
    'GOLD_B': '#F9B775',
    'GOLD_C': '#F0AC5F',
    'GOLD_D': '#E1A158',
    'GOLD_E': '#C78D46',
    'GOLD': '#F0AC5F',
    'RED_A': '#F7A1A3',
    'RED_B': '#FF8080',
    'RED_C': '#FC6255',
    'RED_D': '#E65A4C',
    'RED_E': '#CF5044',
    'PURE_RED': '#FF0000',
    'RED': '#FC6255',
    'MAROON_A': '#ECABC1',
    'MAROON_B': '#EC92AB',
    'MAROON_C': '#C55F73',
    'MAROON_D': '#A24D61',
    'MAROON_E': '#94424F',
    'MAROON': '#C55F73',
    'PURPLE_A': '#CAA3E8',
    'PURPLE_B': '#B189C6',
    'PURPLE_C': '#9A72AC',
    'PURPLE_D': '#715582',
    'PURPLE_E': '#644172',
    'PURPLE': '#9A72AC',
    'PINK': '#D147BD',
    'LIGHT_PINK': '#DC75CD',
    'ORANGE': '#FF862F',
    'LIGHT_BROWN': '#CD853F',
    'DARK_BROWN': '#8B4513',
    'GRAY_BROWN': '#736357',
    'GREY_BROWN': '#736357'
}


class LinearGradient:
    def __init__(self):
        # Multiple types, linear, etc, this must be defined in the preamble/defs region of the svg
        # Hence this must be an object that must be added to the canvas (automatically added maybe)

        # NOTE: SEE: https://www.w3schools.com/graphics/svg_grad_linear.asp

        print('hi')
        print(canvas)
        raise NotImplementedError

    def to_svg(self, time):
        raise NotImplementedError


class RadialGradient:
    def __init__(self):
        # Multiple types, linear, etc, this must be defined in the preamble/defs region of the svg
        # Hence this must be an object that must be added to the canvas (automatically added maybe)
        raise NotImplementedError

    def to_svg(self, time):
        raise NotImplementedError


def color_from_name(name):
    if name not in colors:
        raise ValueError(f'Color {name} is unknown')

    return colors[name]



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QWidget

keymap = {}
for key, value in vars(Qt).items():
    if isinstance(value, Qt.Key):
        keymap[value] = key.partition('_')[2]

modmap = {
    Qt.ControlModifier: keymap[Qt.Key_Control],
    Qt.AltModifier: keymap[Qt.Key_Alt],
    Qt.ShiftModifier: keymap[Qt.Key_Shift],
    Qt.MetaModifier: keymap[Qt.Key_Meta],
    Qt.GroupSwitchModifier: keymap[Qt.Key_AltGr],
    Qt.KeypadModifier: keymap[Qt.Key_NumLock],
    }

def keyevent_to_string(event):
    sequence = []
    for modifier, text in modmap.items():
        if event.modifiers() & modifier:
            sequence.append(text)
    key = keymap.get(event.key(), event.text())
    if key not in sequence:
        sequence.append(key)
    return '+'.join(sequence)



"""
Segments are Lines, Arcs, (Quadratic)BezierCurves
These represent the building blocks of shapes/objects like
    Squares, Circles, TexObjects
Segments do not have styling arguments whereas shapes/objects do.
A path is a collection of segments.

Some thoughts about functionalities are based on "SVG-Morpheus"
See: http://alexk111.github.io/SVG-Morpheus/
See: https://github.com/alexk111/SVG-Morpheus
Copyright (c) 2014 Alex Kaul

This was largely based on "svgpathtools"
See: https://github.com/mathandy/svgpathtools
Copyright (c) 2015 Andrew Allan Port
Copyright (c) 2013-2014 Lennart Regebro

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from abc import ABC
import svgpathtools
import numpy as np
import math
### Local imports
import vectormation.style as style
import vectormation.easings as easings

def sigmoid(x):
  return 1 / (1 + math.exp(-x))

def convert_to_bezier(seg, n_curves=10):
    """Convert svgpathtools segment to Bezier Curve, approximate arcs by n_curves splines"""
    if isinstance(seg, svgpathtools.Line):
        return [Line(start=seg.start, end=seg.end).to_bezier()]
    elif isinstance(seg, svgpathtools.CubicBezier):
        return [CubicBezier(start=seg.start, control1=seg.control1, control2=seg.control2, end=seg.end)]
    elif isinstance(seg, svgpathtools.QuadraticBezier):
        return [QuadraticBezier(start=seg.start, control1=seg.control, end=seg.end)]
    elif isinstance(seg, svgpathtools.Arc):
        # We approximate this arc by n_curves cubic splines, luckily, there is a function for this
        return [convert_to_bezier(bez)[0] for bez in seg.as_cubic_curves(n_curves)]

class Arc(svgpathtools.Arc):
    def to_bezier(self):
        # Split the arc into smaller arcs, then look at
        # https://stackoverflow.com/questions/734076/how-to-best-approximate-a-geometrical-arc-with-a-bezier-curve
        # use self.as_cubic_curves()
        raise NotImplementedError

class CubicBezier(svgpathtools.CubicBezier):
    def to_bezier(self):
        return CubicBezier(start=self.start, control1=self.control1, control2=self.control2, end=self.end)

    def bezier_morph(self, other, easing=easings.smooth):
        """Morphs one bezier curve into the other, returns function of curve with 0<=t<=1"""
        assert isinstance(other, CubicBezier)

        # The curve on [0, 1] that determines the move time is the easing
        cur_s = lambda t: self.start + (other.start - self.start) * easing(t)
        cur_c1 = lambda t: self.control1 + (other.control1 - self.control1) * easing(t)
        cur_c2 = lambda t: self.control2 + (other.control2 - self.control2) * easing(t)
        cur_e = lambda t: self.end + (other.end - self.end) * easing(t)

        return lambda t: CubicBezier(cur_s(t), cur_c1(t), cur_c2(t), cur_e(t))


class QuadraticBezier(svgpathtools.QuadraticBezier):
    def to_bezier(self):
        # https://stackoverflow.com/questions/3162645/convert-a-quadratic-bezier-to-a-cubic-one
        return CubicBezier(
            start=self.start,
            control1=self.start + 2/3*(self.control-self.start),
            control2=self.end + 2/3*(self.control-self.end),
            end=self.end
        )

class Line(svgpathtools.Line):
    def to_bezier(self):
        return CubicBezier(start=self.start, control1=self.start, control2=self.end, end=self.end)

class Path(svgpathtools.Path):
    def adjusted_bbox(self, *transforms):
        """Adjust the bbox of this path to the transforms given in svg-format as strings"""
        if transforms == ('',):
            return self.bbox()
        return self.adjusted_path(*transforms).bbox()

    def adjusted_path(self, *transforms):
        """Adjust the path to the transforms given in svg-format as strings"""
        segs = self._segments

        for transform in transforms[::-1]:
            # Parse the transform into the command and the values
            assert '(' in transform and ')' in transform  # Check for validity
            spl = transform.split('(')
            assert len(spl) == 2
            command, vals = spl
            assert vals[-1] == ')'
            vals = vals[:-1]
            # Split at space or ,
            if vals.count(' ') == 0:
                vals = vals.split(',')
            else:
                vals = vals.split(' ')
            # NOTE: There can be some exceptions to the above rule I believe
            # The below code could result in some errors.
            vals = [float(i) for i in vals]  # See above for details

            # Perform the transformation
            if command == 'scale':
                assert 1 <= len(vals) <= 2  # too little or too much arguments for scale
                x = vals[0]
                if len(vals) == 1:
                    # The y-value is equal to the x-value
                    y = x
                else:
                    y = vals[1]

                # We must scale all of the segments by x, y
                # # The center of the scale is at the center of the current bbox
                # xmin, xmax, ymin, ymax = Path(*segs).bbox()
                # center = (xmin+xmax)/2 + (ymin+ymax)/2 * 1j
                segs = Path(*segs).scaled(x, y, origin=0j)._segments
                # segs = [i.scaled(x, y, origin=0+0j) for i in segs]
                # exit()
            elif command == 'translate':
                assert 1 <= len(vals) <= 2  # too little or too much arguments for translate
                x = vals[0]
                if len(vals) == 1: y = 0
                else: y = vals[1]
                segs = [i.translated(x + y*1j) for i in segs]
            else:
                raise NotImplementedError('Transform not yet implemented')
        return Path(*segs)

class Paths:
    """Is a simple collection of Path objects, e.g. forming two Tex letters"""
    def __init__(self, *args):
        self.paths = []
        self.stylings = []
        for arg in args:
            assert isinstance(arg, tuple)
            assert len(arg) == 2
            assert isinstance(arg[0], svgpathtools.Path)
            assert isinstance(arg[1], style.Styling)
            self.paths.append(arg[0])
            self.stylings.append(arg[1])

    def _morph_prepare(self, other, start_time=0, end_time=1, dist_vs_length=True):
        """Returns the segment pairs in Cubic Bezier form that need be merged
        dist_vs_length is if the subpaths closest in distance or closest in length should be matched"""
        assert isinstance(other, Paths)

        ### General idea
        # Image we have two paths, one being the LaTeX % character and the other a circle.
        # The % has one path with multiple closed subpaths whereas the circle has but one.
        # We circumvent this problem by creating new closed/open paths.
        # All closed subpaths from the starting character are mapped to the closed subpaths
        # of the ending characters, if there are more subpaths to begin with, we add
        # subpaths as closed points with as many segments as needed, if there are less
        # subpaths to begin with we create closed points for the ending characters.
        # The same will be done for the open paths.
        # In this manner we select a matching of closed subpaths before and at the end
        # of the morphing.
        # ADDENDUM: We need to add styling arguments to the shapes that are morphing.
        # If we go from one shape to the other we interpolate the styling between both shapes.
        # However, new closed/open paths do not have any styling. We must thus give these
        # the styling from which shape they came or which they become. The styling is constant.

        # First we list all subpaths of both objects
        closed_subpaths_start = []
        open_subpaths_start = []
        # for path, st in zip(self.paths, self.stylings):
        #     for p in path.continuous_subpaths():
        #         p = Path(*p._segments)
        #         if p.isclosed(): closed_subpaths_start.append((p, st))
        #         else: open_subpaths_start.append((p, st))
        for path, st in zip(self.paths, self.stylings):
            # Note that the subpaths must be non-intersecting
            # NOTE: TODO: loop through all non-intersecting continuous subpaths of the path
            # for the start and end figures.
            if path.iscontinuous() and path.isclosed(): closed_subpaths_start.append((path, st))
            else: open_subpaths_start.append((path, st))

        closed_subpaths_end = []
        open_subpaths_end = []
        for path, st in zip(other.paths, other.stylings):
            # for p in path.continuous_subpaths():
            #     p = Path(*p._segments)
            #     if p.isclosed(): closed_subpaths_end.append((p, st))
            #     else: open_subpaths_end.append((p, st))
            if path.iscontinuous() and path.isclosed(): closed_subpaths_end.append((path, st))
            else: open_subpaths_end.append((path, st))

        # Then we match the closest ones with each other
        subpath_matches = []
        for start, end in [(open_subpaths_start, open_subpaths_end), (closed_subpaths_start, closed_subpaths_end)]:
            mat_dist = np.zeros((len(start), len(end)))
            for i, (path, path_style) in enumerate(start):
                if dist_vs_length:
                    transforms = path_style.transform_style(start_time)
                    transforms = transforms[1:].split(' ')
                    xmin, xmax, ymin, ymax = path.adjusted_bbox(*transforms)
                    path_center = ((xmin+xmax)/2, (ymin+ymax)/2)

                for j, (match, match_style) in enumerate(end):
                    if dist_vs_length:
                        # NOTE: TODO: cache these adjusted bbox results.
                        transforms = match_style.transform_style(end_time)
                        transforms = transforms[1:].split(' ')
                        xmin, xmax, ymin, ymax = match.adjusted_bbox(*transforms)
                        match_center = ((xmin+xmax)/2, (ymin+ymax)/2)

                        dist = ((match_center[0]-path_center[0])**2 + (match_center[1]-path_center[1])**2)**(1/2)

                        mat_dist[i,j] = dist
                    else:
                        mat_dist[i,j] = abs(path.length() / match.length() - 1)

            ls, le = len(start), len(end)
            matches = []
            # Pre-compute the largest value in the matrix
            for _ in range(min(ls, le)):
                match_idx = np.unravel_index(np.nanargmin(mat_dist, axis=None), mat_dist.shape)
                # Remove objects from matrix
                # mat_dist = np.delete(mat_dist, match_idx[0], axis=0)
                # mat_dist = np.delete(mat_dist, match_idx[1], axis=1)
                # Set objects in matrix to zero to prevent from making double matches (-1 < min(mat_dist))
                mat_dist[match_idx[0],:] = np.nan
                mat_dist[:,match_idx[1]] = np.nan
                matches.append(match_idx)

            # We now need to add the new subpaths for the matches
            if ls >= le:  # We must create extra open ending points
                for idx in {i for i in range(ls)} - {m[0] for m in matches}:
                    path, path_style = start[idx]
                    subpath_matches.append((start[idx], (Path(), path_style)))  # The same styling
            else:  # We must create extra open starting points
                for idx in {i for i in range(le)} - {m[1] for m in matches}:
                    path, path_style = end[idx]
                    subpath_matches.append(((Path(), path_style), end[idx]))  # The same styling
            # Add all correctly matched pairs
            for i, j in matches:
                subpath_matches.append((start[i], end[j]))

        # Now that all subpaths are matched we match all segments in the subpaths
        subpath_segment_pairs = []
        for (path_from, style_from), (path_to, style_to) in subpath_matches:
            segment_pairs = []
            # We have only implemented the morphing of cubic bezier thus we need to convert
            # all shapes to these (an arc is split into multiple smaller arcs before converting).
            segs_from = []
            for seg in path_from:
                segs_from += convert_to_bezier(seg)
            segs_to = []
            for seg in path_to:
                segs_to += convert_to_bezier(seg)

            lf, lt = len(segs_from), len(segs_to)
            if min(lf, lt) == 0:
                # We have to create this path from nothing
                if lf == 0:
                    xmin, xmax, ymin, ymax = path_to.bbox()
                    center = (xmin+xmax)/2 + (ymin+ymax)/2 * 1j
                    for i in segs_to:
                        segment_pairs.append((Line(start=center, end=center).to_bezier(), i))
                    subpath_segment_pairs.append((segment_pairs, style_from, style_to))
                    continue
                else:
                    xmin, xmax, ymin, ymax = path_from.bbox()
                    center = (xmin+xmax)/2 + (ymin+ymax)/2 * 1j
                    for i in segs_from:
                        segment_pairs.append((i, Line(start=center, end=center).to_bezier()))
                    subpath_segment_pairs.append((segment_pairs, style_from, style_to))
                    continue

            # Split the longest lines in half
            while len(segs_from) < len(segs_to):
                max_idx = max(enumerate(segs_from), key=lambda x: x[1].length())[0]
                new1, new2 = segs_from[max_idx].split(0.5)
                new1, new2 = convert_to_bezier(new1), convert_to_bezier(new2)
                segs_from = segs_from[:max_idx] + new1 + new2 + segs_from[max_idx+1:]
            # Split the longest lines in half
            while len(segs_to) < len(segs_from):
                max_idx = max(enumerate(segs_to), key=lambda x: x[1].length())[0]
                new1, new2 = segs_to[max_idx].split(0.5)
                new1, new2 = convert_to_bezier(new1), convert_to_bezier(new2)
                segs_to = segs_to[:max_idx] + new1 + new2 + segs_to[max_idx+1:]

            segment_pairs += list(zip(segs_from, segs_to))
            subpath_segment_pairs.append((segment_pairs, style_from, style_to))

        # Save all information
        self.subpath_segment_pairs = subpath_segment_pairs

        return subpath_segment_pairs


    def morph(self, other, start_time=0, end_time=1, easing=easings.smooth):
        # Store the morph prepare output (by id of other) so processing only happens once
        if hasattr(self, 'segment_pairs'):
            subpath_segment_pairs = self.subpath_segment_pairs
        else:
            subpath_segment_pairs = self._morph_prepare(other, start_time, end_time, dist_vs_length=True)

        objects = []
        for segment_pairs, style_from, style_to in subpath_segment_pairs:
            path = []
            for f, t in segment_pairs:
                seg_func = f.bezier_morph(t, easing=easing)
                path.append(seg_func)

            # Define the subpath we need to draw as a function of time (with 0<=t<=1)
            def f(path):  # We need to avoid a direct lambda function because of locality of path
                return lambda t: Path(*[s(t) for s in path]).d()
            object = (f(path), style_from, style_to)
            # The object has styling style_from and style_to, we need to interpolate
            # these style objects and add the interpolation to the path data
            # This is done in the __init__ of the Morph class
            objects.append(object)

        return objects



if __name__ == '__main__':
    # p1 = svgpathtools.Path('M0,1 L100,200 L400,200 Z')
    # p2 = svgpathtools.Path('M1000,1 L200,200 L900,500 Z')
    p1 = svgpathtools.Path('M0,1 L100,200 L400,200 Z')
    # p2 = svgpathtools.Path('M0,2 L300,300 C100,200 110,210 340,230 A100,200 200,300 300,400 400')
    p2 = svgpathtools.Path('M0,2 L300,300 C100,200 110,210 340,230')
    p3 = svgpathtools.Path('M0,1 L100,200 L400,200 Z M0,2 L100,200 L400,200 Z M0,3 L20,30 M0,4 L40,50')
    # p1.morph(p2)

    f = Paths(p1, p2).morph(Paths(p3))
    print(f(0))
    # TODO: add close parameter Z to svg path
    # TODO: add morphobject in script
    # TODO: add rotation (360 degrees) as transform to that morphobject
    # TODO: Return as same collection of subpaths for rotation?



import sys
from PyQt5.QtCore import QFile, QSize, Qt, QEvent
from PyQt5.QtGui import QBrush, QColor, QImage, QPainter, QPixmap, QPen
from PyQt5.QtWidgets import (QActionGroup, QApplication, QFileDialog,
        QGraphicsItem, QGraphicsRectItem, QGraphicsScene, QGraphicsView,
        QMainWindow, QMenu, QMessageBox, QWidget, QProgressBar,
        QHBoxLayout, QMenuBar, QPushButton, QLabel)
from PyQt5.QtOpenGL import QGL, QGLFormat, QGLWidget
from PyQt5.QtSvg import QGraphicsSvgItem


class MainWindow(QMainWindow):
    def __init__(self, display, use_menubar=True):
        super(MainWindow, self).__init__()
        # Set the viewer and the placeholder for the path to the displayed svg
        self.view = SvgView(self, display)
        self.currentPath = ''

        # Add a default size for floating windows
        self.resize(800, 600)

        # Add mouse tracking
        self.setMouseTracking(True)

        if use_menubar:
            # Define the layout of the top bar
            self.layout = QHBoxLayout()
            self.layout.setContentsMargins(0, 0, 0, 0)

            ### Add the renderers to an option menu
            rendererMenu = QMenu("&Renderer", self)
            self.nativeAction = rendererMenu.addAction("&Native")
            self.nativeAction.setCheckable(True)
            self.nativeAction.setChecked(True)

            if QGLFormat.hasOpenGL():
                self.glAction = rendererMenu.addAction("&OpenGL")
                self.glAction.setCheckable(True)

            self.imageAction = rendererMenu.addAction("&Image")
            self.imageAction.setCheckable(True)

            if QGLFormat.hasOpenGL():
                rendererMenu.addSeparator()
                self.highQualityAntialiasingAction = rendererMenu.addAction("&High Quality Antialiasing")
                self.highQualityAntialiasingAction.setEnabled(False)
                self.highQualityAntialiasingAction.setCheckable(True)
                self.highQualityAntialiasingAction.setChecked(False)
                self.highQualityAntialiasingAction.toggled.connect(self.view.setHighQualityAntialiasing)

            rendererGroup = QActionGroup(self)
            rendererGroup.addAction(self.nativeAction)

            if QGLFormat.hasOpenGL():
                rendererGroup.addAction(self.glAction)

            rendererGroup.addAction(self.imageAction)
            rendererGroup.triggered.connect(self.setRenderer)

            # Add renderer options to menubar
            self.menu_bar = QMenuBar()
            self.menu_bar.addMenu(rendererMenu)

            # Add start and stop buttons for the animation
            start_button = QPushButton('&Restart')
            start_button.clicked.connect(self.start_anim)
            stop_button = QPushButton('&Stop')
            stop_button.clicked.connect(self.stop_anim)

            # Add a progress bar for displaying the frame count
            self.progress_label = QLabel('Frame:')
            self.progress = QProgressBar(self)
            self.progress.setMaximumWidth(2000)
            # self.progress.setRange(0, 100)
            # self.progress.setValue(0)
            self.progress.setFormat(f'{0}/??')

            # Add x and y positions of mouse
            self.xpos = QLabel(f'x=?')
            self.ypos = QLabel(f'y=?')

            # Add the widgets to the layout
            self.menu_bar.setMaximumSize(rendererMenu.size())
            self.layout.addWidget(self.menu_bar)
            self.layout.addStretch()
            self.layout.addWidget(self.progress_label)
            self.layout.addWidget(self.progress)
            self.layout.addStretch()
            self.layout.addWidget(self.xpos)
            self.layout.addWidget(self.ypos)
            self.layout.addWidget(start_button)
            self.layout.addWidget(stop_button)
            # Add layout to a bar
            widget = QWidget(self)
            self.setMenuWidget(widget)
            widget.setLayout(self.layout)

        # Make the SvgView widget the displayed widget
        self.setCentralWidget(self.view)
        self.setWindowTitle("SVG Viewer")

    def start_anim(self):
        """Restart the animation"""
        canvas = self.view.display
        canvas.time = canvas.start_anim
        canvas.frame_count = 0
        return True

    def stop_anim(self):
        """Stop animating..."""
        self.view.display.animate = 1 - self.view.display.animate
        return True

    def openFile(self, path=None):
        """Opens and checks the svg file's existence"""
        if path:
            svg_file = QFile(path)
            if not svg_file.exists():
                QMessageBox.critical(self, "Open SVG File",
                        "Could not open file '%s'." % path)
                return

            self.view.openFile(svg_file)

            if not path.startswith(':/'):
                self.currentPath = path
                self.setWindowTitle("%s - SVGViewer" % self.currentPath)

    def setRenderer(self, action):
        """Change the renderer"""
        if QGLFormat.hasOpenGL():
            self.highQualityAntialiasingAction.setEnabled(False)

        if action == self.nativeAction:
            self.view.setRenderer(SvgView.Native)
        elif action == self.glAction:
            if QGLFormat.hasOpenGL():
                self.highQualityAntialiasingAction.setEnabled(True)
                self.view.setRenderer(SvgView.OpenGL)
        elif action == self.imageAction:
            self.view.setRenderer(SvgView.Image)

    def closeEvent(self, event):
        print('Detecting a window-kill, closing...')
        sys.exit()

    """Sent all info to the animation"""
    def eventFilter(self, qobject, qevent):
        return self.view.eventFilter(event)

    def resizeEvent(self, event):
        self.view.resizeEvent(event)

    def keyPressEvent(self, event):
        self.view.keyPressEvent(event)

    def mousePressEvent(self, event):
        self.view.mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self.view.mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):
        self.view.mouseMoveEvent(event)


class SvgView(QGraphicsView):
    Native, OpenGL, Image = range(3)

    def __init__(self, mainwindow, display, parent=None):
        super(SvgView, self).__init__(parent)

        self.display = display
        self.renderer = SvgView.Native
        self.svgItem = None
        self.windowBorder = None
        self.mainwindow = mainwindow
        self.image = QImage()

        # Add mouse tracking
        self.setMouseTracking(True)

        # Disable scrolling bars
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.setScene(QGraphicsScene(self))

        # Prepare background check-board pattern.
        tilePixmap = QPixmap(64, 64)
        tilePixmap.fill(Qt.white)
        tilePainter = QPainter(tilePixmap)
        color = QColor(220, 220, 220)
        tilePainter.fillRect(0, 0, 32, 32, color)
        tilePainter.fillRect(32, 32, 32, 32, color)
        tilePainter.end()

        self.setBackgroundBrush(QBrush(tilePixmap))
        # self.setBackgroundBrush(QBrush(Qt.black))


    def openFile(self, svg_file):
        """Draws the contents of an svg file"""
        if not svg_file.exists():
            return

        s = self.scene()
        s.clear()
        self.resetTransform()

        # Draw the svgviewer
        self.svgItem = QGraphicsSvgItem(svg_file.fileName())
        self.svgItem.setFlags(QGraphicsItem.ItemClipsToShape)
        self.svgItem.setCacheMode(QGraphicsItem.NoCache)
        self.svgItem.setZValue(0)

        # Get the svgviewer's bounding rectangle (adjusted for the width of the windowBorder)
        sw = 2
        bounding = self.svgItem.boundingRect()
        self.windowBorder = QGraphicsRectItem(bounding)

        # Draw a dashed border around the viewer
        outline = QPen(Qt.blue, sw)
        self.windowBorder.setPen(outline)
        self.windowBorder.setBrush(QBrush(Qt.NoBrush))
        self.windowBorder.setZValue(1)

        s.addItem(self.svgItem)
        s.addItem(self.windowBorder)

        # Resize the window to fit the drawings
        self.resizeEvent(0)


    def setRenderer(self, renderer):
        """Change the renderer"""
        self.renderer = renderer

        if self.renderer == SvgView.OpenGL:
            if QGLFormat.hasOpenGL():
                self.setViewport(QGLWidget(QGLFormat(QGL.SampleBuffers)))
        else:
            self.setViewport(QWidget())

    def setHighQualityAntialiasing(self, highQualityAntialiasing):
        if QGLFormat.hasOpenGL():
            self.setRenderHint(QPainter.HighQualityAntialiasing,
                    highQualityAntialiasing)


    def paintEvent(self, event):
        """Redraws the canvas"""
        if self.renderer == SvgView.Image:
            if self.image.size() != self.viewport().size():
                self.image = QImage(self.viewport().size(),
                        QImage.Format_ARGB32_Premultiplied)

            imagePainter = QPainter(self.image)
            QGraphicsView.render(self, imagePainter)
            imagePainter.end()

            p = QPainter(self.viewport())
            p.drawImage(0, 0, self.image)
        else:
            super(SvgView, self).paintEvent(event)

    """Sent all noteworthy events to the display"""
    def eventFilter(self, _, qevent):
        qtype = qevent.type()

        if qtype in [QEvent.Timer, QEvent.Paint]:
            # Nothing important happens
            qevent.accept()
            return True
        else:
            # Sent the event
            self.display.event(qevent)
        qevent.accept()
        return True

    def keyPressEvent(self, event):
        self.display.event(event)

    def mousePressEvent(self, event):
        self.display.event(event)

    def mouseReleaseEvent(self, event):
        self.display.event(event)

    def mouseMoveEvent(self, event):
        self.display.event(event)

    def resizeEvent(self, _):
        """Rescale the window to fit the svg-viewer"""
        if self.svgItem != None:
            w, h = self.width()+4, self.height()+4
            bounding = self.svgItem.boundingRect()
            svgw, svgh = bounding.width(), bounding.height()
            sc = min((w/svgw, 'w'), (h/svgh, 'h'))

            self.svgItem.setScale(sc[0])
            self.windowBorder.setScale(sc[0])
            if sc[1] == 'h':
                self.svgItem.setPos((w-sc[0]*svgw)//2, 0)
                self.windowBorder.setPos((w-sc[0]*svgw)//2, 0)
            else:
                self.svgItem.setPos(0, (h-sc[0]*svgh)//2)
                self.windowBorder.setPos(0, (h-sc[0]*svgh)//2)
            self.scene().setSceneRect(0, 0, w, h)


class SVGViewer:
    """A simple Qt-powered SVGViewer that can open files untill destruction"""
    def __init__(self, display):
        self.app = QApplication(sys.argv)

        self.window = MainWindow(display)
        self.window.show()

    def view(self, filename, wait_till=None):
        self.window.openFile(filename)
        if wait_till is not None:
            from PyQt5 import QtTest
            import datetime
            to_wait = wait_till - datetime.datetime.now()
            secs = to_wait.total_seconds()
            QtTest.QTest.qWait(int(secs*1000))

    def destroy(self):
        # sys.exit(self.app.exec_())
        sys.exit(self.app.exec())



"""
NOTE: these easings/rate functions are copied from
https://github.com/ManimCommunity/manim/blob/main/manim/utils/rate_functions.py

MIT License

Copyright (c) 2018 3Blue1Brown LLC
Copyright (c) 2021, the Manim Community Developers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import typing
from functools import wraps
import numpy as np


# This is a decorator that makes sure any function it's used on will
# return 0 if t<0 and 1 if t>1.
def unit_interval(function):
    @wraps(function)
    def wrapper(t, *args, **kwargs):
        if 0 <= t <= 1:
            return function(t, *args, **kwargs)
        elif t < 0:
            return 0
        else:
            return 1

    return wrapper


# This is a decorator that makes sure any function it's used on will
# return 0 if t<0 or t>1.
def zero(function):
    @wraps(function)
    def wrapper(t, *args, **kwargs):
        if 0 <= t <= 1:
            return function(t, *args, **kwargs)
        else:
            return 0

    return wrapper

def sigmoid(x: float) -> float:
    r"""Returns the output of the logistic function.
    The logistic function, a common example of a sigmoid function, is defined
    as :math:`\frac{1}{1 + e^{-x}}`.
    References
    ----------
    - https://en.wikipedia.org/wiki/Sigmoid_function
    - https://en.wikipedia.org/wiki/Logistic_function
    """
    return 1.0 / (1 + np.exp(-x))

@unit_interval
def linear(t: float) -> float:
    return t


@unit_interval
def smooth(t: float, inflection: float = 10.0) -> float:
    error = sigmoid(-inflection / 2)
    return min(
        max((sigmoid(inflection * (t - 0.5)) - error) / (1 - 2 * error), 0),
        1,
    )


@unit_interval
def rush_into(t: float, inflection: float = 10.0) -> float:
    return 2 * smooth(t / 2.0, inflection)


@unit_interval
def rush_from(t: float, inflection: float = 10.0) -> float:
    return 2 * smooth(t / 2.0 + 0.5, inflection) - 1


@unit_interval
def slow_into(t: float) -> float:
    return np.sqrt(1 - (1 - t) * (1 - t))


@unit_interval
def double_smooth(t: float) -> float:
    if t < 0.5:
        return 0.5 * smooth(2 * t)
    else:
        return 0.5 * (1 + smooth(2 * t - 1))


@zero
def there_and_back(t: float, inflection: float = 10.0) -> float:
    new_t = 2 * t if t < 0.5 else 2 * (1 - t)
    return smooth(new_t, inflection)


@zero
def there_and_back_with_pause(t: float, pause_ratio: float = 1.0 / 3) -> float:
    a = 1.0 / pause_ratio
    if t < 0.5 - pause_ratio / 2:
        return smooth(a * t)
    elif t < 0.5 + pause_ratio / 2:
        return 1
    else:
        return smooth(a - a * t)

def not_quite_there(
    func: typing.Callable[[float], float] = smooth,
    proportion: float = 0.7,
) -> typing.Callable[[float], float]:
    def result(t):
        return proportion * func(t)

    return result


@zero
def wiggle(t: float, wiggles: float = 2) -> float:
    return there_and_back(t) * np.sin(wiggles * np.pi * t)


def squish_rate_func(
    func: typing.Callable[[float], float],
    a: float = 0.4,
    b: float = 0.6,
) -> typing.Callable[[float], float]:
    def result(t):
        if a == b:
            return a

        if t < a:
            return func(0)
        elif t > b:
            return func(1)
        else:
            return func((t - a) / (b - a))

    return result


# Stylistically, should this take parameters (with default values)?
# Ultimately, the functionality is entirely subsumed by squish_rate_func,
# but it may be useful to have a nice name for with nice default params for
# "lingering", different from squish_rate_func's default params


@unit_interval
def lingering(t: float) -> float:
    return squish_rate_func(lambda t: t, 0, 0.8)(t)


@unit_interval
def exponential_decay(t: float, half_life: float = 0.1) -> float:
    # The half-life should be rather small to minimize
    # the cut-off error at the end
    return 1 - np.exp(-t / half_life)


@unit_interval
def ease_in_sine(t: float) -> float:
    return 1 - np.cos((t * np.pi) / 2)


@unit_interval
def ease_out_sine(t: float) -> float:
    return np.sin((t * np.pi) / 2)


@unit_interval
def ease_in_out_sine(t: float) -> float:
    return -(np.cos(np.pi * t) - 1) / 2


@unit_interval
def ease_in_quad(t: float) -> float:
    return t * t


@unit_interval
def ease_out_quad(t: float) -> float:
    return 1 - (1 - t) * (1 - t)


@unit_interval
def ease_in_out_quad(t: float) -> float:
    return 2 * t * t if t < 0.5 else 1 - pow(-2 * t + 2, 2) / 2


@unit_interval
def ease_in_cubic(t: float) -> float:
    return t * t * t


@unit_interval
def ease_out_cubic(t: float) -> float:
    return 1 - pow(1 - t, 3)


@unit_interval
def ease_in_out_cubic(t: float) -> float:
    return 4 * t * t * t if t < 0.5 else 1 - pow(-2 * t + 2, 3) / 2


@unit_interval
def ease_in_quart(t: float) -> float:
    return t * t * t * t


@unit_interval
def ease_out_quart(t: float) -> float:
    return 1 - pow(1 - t, 4)


@unit_interval
def ease_in_out_quart(t: float) -> float:
    return 8 * t * t * t * t if t < 0.5 else 1 - pow(-2 * t + 2, 4) / 2


@unit_interval
def ease_in_quint(t: float) -> float:
    return t * t * t * t * t


@unit_interval
def ease_out_quint(t: float) -> float:
    return 1 - pow(1 - t, 5)


@unit_interval
def ease_in_out_quint(t: float) -> float:
    return 16 * t * t * t * t * t if t < 0.5 else 1 - pow(-2 * t + 2, 5) / 2


@unit_interval
def ease_in_expo(t: float) -> float:
    return 0 if t == 0 else pow(2, 10 * t - 10)


@unit_interval
def ease_out_expo(t: float) -> float:
    return 1 if t == 1 else 1 - pow(2, -10 * t)


@unit_interval
def ease_in_out_expo(t: float) -> float:
    if t == 0:
        return 0
    elif t == 1:
        return 1
    elif t < 0.5:
        return pow(2, 20 * t - 10) / 2
    else:
        return (2 - pow(2, -20 * t + 10)) / 2


@unit_interval
def ease_in_circ(t: float) -> float:
    return 1 - np.sqrt(1 - pow(t, 2))


@unit_interval
def ease_out_circ(t: float) -> float:
    return np.sqrt(1 - pow(t - 1, 2))


@unit_interval
def ease_in_out_circ(t: float) -> float:
    return (
        (1 - np.sqrt(1 - pow(2 * t, 2))) / 2
        if t < 0.5
        else (np.sqrt(1 - pow(-2 * t + 2, 2)) + 1) / 2
    )


@unit_interval
def ease_in_back(t: float) -> float:
    c1 = 1.70158
    c3 = c1 + 1
    return c3 * t * t * t - c1 * t * t


@unit_interval
def ease_out_back(t: float) -> float:
    c1 = 1.70158
    c3 = c1 + 1
    return 1 + c3 * pow(t - 1, 3) + c1 * pow(t - 1, 2)


@unit_interval
def ease_in_out_back(t: float) -> float:
    c1 = 1.70158
    c2 = c1 * 1.525
    return (
        (pow(2 * t, 2) * ((c2 + 1) * 2 * t - c2)) / 2
        if t < 0.5
        else (pow(2 * t - 2, 2) * ((c2 + 1) * (t * 2 - 2) + c2) + 2) / 2
    )


@unit_interval
def ease_in_elastic(t: float) -> float:
    c4 = (2 * np.pi) / 3
    if t == 0:
        return 0
    elif t == 1:
        return 1
    else:
        return -pow(2, 10 * t - 10) * np.sin((t * 10 - 10.75) * c4)


@unit_interval
def ease_out_elastic(t: float) -> float:
    c4 = (2 * np.pi) / 3
    if t == 0:
        return 0
    elif t == 1:
        return 1
    else:
        return pow(2, -10 * t) * np.sin((t * 10 - 0.75) * c4) + 1


@unit_interval
def ease_in_out_elastic(t: float) -> float:
    c5 = (2 * np.pi) / 4.5
    if t == 0:
        return 0
    elif t == 1:
        return 1
    elif t < 0.5:
        return -(pow(2, 20 * t - 10) * np.sin((20 * t - 11.125) * c5)) / 2
    else:
        return (pow(2, -20 * t + 10) * np.sin((20 * t - 11.125) * c5)) / 2 + 1


@unit_interval
def ease_in_bounce(t: float) -> float:
    return 1 - ease_out_bounce(1 - t)


@unit_interval
def ease_out_bounce(t: float) -> float:
    n1 = 7.5625
    d1 = 2.75

    if t < 1 / d1:
        return n1 * t * t
    elif t < 2 / d1:
        return n1 * (t - 1.5 / d1) * (t - 1.5 / d1) + 0.75
    elif t < 2.5 / d1:
        return n1 * (t - 2.25 / d1) * (t - 2.25 / d1) + 0.9375
    else:
        return n1 * (t - 2.625 / d1) * (t - 2.625 / d1) + 0.984375


@unit_interval
def ease_in_out_bounce(t: float) -> float:
    if t < 0.5:
        return (1 - ease_out_bounce(1 - 2 * t)) / 2
    else:
        return (1 + ease_out_bounce(2 * t - 1)) / 2



"""Here we define all objects like squares, texobjects, morphobjects, ...
First we create a general object that defines the necessary functions and a fadein-fadeout function
"""
import sys
from abc import ABC, abstractmethod
from copy import copy, deepcopy
from bs4 import BeautifulSoup
from time import sleep
from svgpathtools import parse_path  # Write code yourself??? only bbox needed, NOTE: TODO
import math
import datetime

# Local imports
import vectormation.easings as easings
import vectormation.attributes as attributes
from vectormation.tex_file_writing import tex_content_to_svg_file
from vectormation.tex_file_writing import get_characters
from vectormation.utils import keyevent_to_string
from vectormation.window import SVGViewer  # NOTE: TODO is this import necessary or can we disable it
import vectormation.style as style
import vectormation.morphing as morphing

# MARKER: Above code is not checked, restucture files; NOTE: TODO

class VectorMathAnim:
    """Video where we can ask a frame at a certain time"""
    def __init__(self, save_dir, width=1000, height=1000):
        self.save_dir = save_dir  # Directory to save frames in
        self.filename = f'{save_dir}/gen.svg'
        self.width = width
        self.height = height
        self.viewbox = (0, 0, width, height)
        self.objects = {}  # {id(object): object}
        self.time = 0  # Time of the current video
        self.dt = 0
        self.start_anim = None
        self.end_anim = None
        self.animate = None
        self.background = None
        self.svgviewer = None

        # We need to set the save directory globally
        global save_directory
        save_directory = save_dir

    def set_background(self, creation=0, z=-1, **styling):
        """Sets the background of the animation (otherwise no background is added)"""
        if self.background != None:
            # The background already existed, discarding...
            del self.objects[id(self.background)]
        # Process the styling keyword arguments with a default black background
        st = style.Styling(styling, creation=creation, stroke_width=0)
        self.background = Rectangle(self.width, self.height, x=0, y=0, rx=0, ry=0, creation=creation, z=z, **st.kwargs())
        self.add_objects(self.background)

    def add_objects(self, *args):
        """Adding to a collection of all objects that must be displayed at some time"""
        for obj in args:
            obj_id = id(obj)
            self.objects[obj_id] = obj

    def write_frame(self, time=None, filename=None):
        """This combines all the svgs of objects alive at the canvas time"""
        if time == None:
            time = self.time  # The canvas time
        if filename == None:
            filename = self.filename

        # Write the svg data to the file
        with open(filename, 'w') as s:
            s.write("<?xml version='1.0' encoding='UTF-8'?>\n")
            header = f"<svg version='1.1' xmlns='http://www.w3.org/2000/svg' " \
                     f"xmlns:xlink='http://www.w3.org/1999/xlink' " \
                     f"width='{self.width}' height='{self.height}' " \
                     f"viewBox='{' '.join(str(i) for i in self.viewbox)}'>\n"
            s.write(header)
            # Sorting based on the z property/draw order
            lst = []
            for obj in self.objects.values():
                if obj.show.at_time(time):
                    if not hasattr(obj, 'z'):
                        lst.append((0, obj))
                    else:
                        lst.append((obj.z.at_time(time), obj))

            # Add the objects
            for _, obj in sorted(lst, key=lambda x: x[0]):
                s.write(obj.to_svg(time) + '\n')

            # Close the header
            s.write("</svg>")

    def event(self, event):
        """Code to process any events from the animation window"""
        if event.__class__.__name__ == 'QMouseEvent':
            ### Display the mouse position in the top labels
            w, h = self.svgviewer.window.width(), self.svgviewer.window.height()
            svg_part = self.svgviewer.window.view.svgItem
            bounding = svg_part.boundingRect()
            bw, bh = bounding.width(), bounding.height()
            scale = svg_part.scale()
            w, h = bw*scale, bh*scale
            x, y = svg_part.pos().x(), svg_part.pos().y()
            # The x, y, w, h is the rectangle around the svgview part
            map_x, map_y = event.x(), event.y()
            # The relative position of the mouse
            rel_x, rel_y = (map_x-x)/w, (map_y-y)/h
            # The coordinates of the mouse with respect to the viewer
            view_x = rel_x * self.viewbox[2] + self.viewbox[0]
            view_y = rel_y * self.viewbox[3] + self.viewbox[1]

            # Add this information to the viewer
            self.svgviewer.window.xpos.setText(f'x={view_x:#.2f}')
            self.svgviewer.window.ypos.setText(f'y={view_y:#.2f}')
            # NOTE: TODO: Dragging does what?
        elif event.__class__.__name__ == 'QWheelEvent':
            ### Scrolling motion
            zoom = pow(1.2, event.pixelDelta().y() / 240.0)
            pos = event.x(), event.y()
            v = self.viewbox
            # Mouse position relative to the screen
            svg_part = self.svgviewer.window.view.svgItem
            rect = svg_part.boundingRect()
            rel_x = (pos[0]-svg_part.pos().x())/rect.width()
            rel_y = (pos[1]-svg_part.pos().y())/rect.height()
            rel_x, rel_y = max(min(rel_x, 1), 0), max(min(rel_y, 1), 0)
            # Change the viewbox
            new_width, new_height = v[2]/zoom, v[3]/zoom
            dw, dh = v[2]-new_width, v[3]-new_height
            self.viewbox = (v[0]+rel_x*dw, v[1]+rel_y*dh, new_width, new_height)
            self.redraw()
        elif event.__class__.__name__ == 'QKeyEvent':
            ### Actions for keypresses
            char = keyevent_to_string(event)
            if char == 'Q':
                print('Quitting...')
                sys.exit()
            elif char == 'R':
                # Restarting the animation
                self.time = self.start_anim
                self.frame_count = 0
            elif char == 'F':
                # Resetting window zoom
                self.viewbox = (0, 0, self.width, self.height)
                self.redraw()
            elif char == 'P':
                # Pausing (or toggling pause) of window
                self.animate = 1 - self.animate
                self.redraw()
        else:
            # Discard other events like window open events
            pass

    def redraw(self):
        """Force the svg viewer to redraw"""
        if self.svgviewer is None:
            raise BaseException('Cannot redraw, no svg viewer')
        self.write_frame()
        self.svgviewer.view(self.filename)

    def standard_display(self, start_time=0, end_time=None, fps=60):
        """Views the svg using the svg viewer
        renders with accuracy of about a millisecond on time
        If start_time < 0, we go that many seconds before the end"""
        self.svgviewer = SVGViewer(self)
        if end_time == None:
            # Animate to the last available frame, first we must find it
            end_time = max(obj.last_change for obj in self.objects.values())
            print(f'Found that the ending time is {end_time}.')
        self.end_anim = end_time
        # Set the starting time of the animation
        if start_time < 0:
            self.start_anim = end_time - start_time
        else:
            self.start_anim = start_time
        self.animate = True

        # Start animating at the start time and increase with dt
        # untill the end_time is reached
        self.time = start_time
        self.frame_count = 0
        self.dt = 1/fps
        while True:
            # Continue displaying and redrawing forever
            wait_till = datetime.datetime.now() + datetime.timedelta(seconds=self.dt)
            self.write_frame()
            self.svgviewer.view(self.filename, wait_till)

            if self.frame_count <= self.end_anim * fps:
                # Change progressbar
                prog = self.svgviewer.window.progress
                if end_time == 0:
                    prog.setValue(100)
                else:
                    prog.setValue(int(self.frame_count / (end_time*fps)*100))
                prog.setFormat(f'{self.frame_count}/{round(end_time*fps)}')

            # Increase the time so that we can draw the next frame
            if self.animate:
                self.time += self.dt
                self.time = min(self.time, end_time)  # to avoid rounding error affecting the display
                self.frame_count += 1


class VObject(ABC):  # Vector Object
    """This is the motherclass for all vector objects
    Having creation and destroying function implemented.
    Since an image is not a vector object this will not be a child
    (thus the fadein function and fadeout function will not be the
     same as for vector objects)
    """
    @abstractmethod
    def __init__(self, creation=0, z=0):
        # If we need to show the object
        self.show = attributes.Real(creation, True)
        # Specify the draw order
        self.z = attributes.Real(creation, z)

    @abstractmethod
    def to_svg(self, time):
        """Get the svg-specifyer of the object at this time"""
        pass

    @abstractmethod
    def last_change(self):
        """Get the time of the last change of this shape, the max of all attributes"""
        pass

    @abstractmethod
    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        """Shift the object dx and dy (without changing the transformation styling preferrably)
        If end_time is not None we animate the movement from start_time to end_time"""
        pass

    @abstractmethod
    def path(self, time):
        """Get the SVG-path-specifyer of the object (unadjusted for its transforms)"""
        pass

    def bbox(self, time):
        """Get the bounding rectangle in (xmin, ymin, width, height) at a certain time"""
        # Get the path specifyer of the object
        path = self.path(time)
        if not hasattr(self, 'styling'):
            # We do not need to adjust for the transform part of styling
            xmin, xmax, ymin, ymax = morphing.Path(path).bbox()
        else:
            transforms = self.styling.transform_style(time)
            if transforms == '':
                # No transforms to be done
                xmin, xmax, ymin, ymax = morphing.Path(path).bbox()
                return (xmin, ymin, xmax-xmin, ymax-ymin)
            transforms = transforms[1:].split(' ')
            xmin, xmax, ymin, ymax = morphing.Path(path).adjusted_bbox(*transforms)
        return (xmin, ymin, xmax-xmin, ymax-ymin)

    def center_to_pos(self, posx=500, posy=500, start_time=0, end_time=None, easing=easings.smooth):
        """Shifts the center to pos, and animates this from start_time to end_time
        Shifting is done with respect to the position at start_time."""
        xmin, ymin, width, height = self.bbox(start_time)
        dx, dy = posx-(xmin+width/2), posy-(ymin+height/2)
        self.shift(dx=dx, dy=dy, start_time=start_time, end_time=end_time, easing=easing)

    def brect(self, time=0, rx=0, ry=0, dpos=5, follow=True):
        """Returns the bounding rectangle (or surrounding rectangle) adjusted with dpos outward
        If time is specified creates the rectangle from time onward
        If follow = True, the rectangle will follow the object."""
        if not follow:  # Get the bounding rectangle once
            x, y, width, height = self.bbox(time)
            return Rectangle(
                width+2*dpos, height+2*dpos, x-dpos, y-dpos, rx=rx, ry=ry, creation=time,
                fill_opacity=0, stroke_opacity=1, stroke='#ff0', stroke_width=2
            )
        else:
            rect = Rectangle(width=0, height=0, rx=rx, ry=ry, creation=time,
                             fill_opacity=0, stroke_opacity=1, stroke='#ff0', stroke_width=2)
            rect.x.set_onward(time, lambda t: self.bbox(t)[0] - dpos)
            rect.y.set_onward(time, lambda t: self.bbox(t)[1] - dpos)
            rect.width.set_onward(time, lambda t: self.bbox(t)[2] + 2*dpos)
            rect.height.set_onward(time, lambda t: self.bbox(t)[3] + 2*dpos)
            return rect

    def fadein(self, start=0, end=1, change_existence=True, easing=easings.linear):
        """This object will get to the original opacity at the end starting from none
        The show argument before start will be set to none if change_existence, this is a creating function"""
        # This creates the object, we can thus not see it before this command
        if change_existence:
            self.show.set_onward(0, False)
            self.show.set_onward(start, True)

        end_val = self.styling.opacity.at_time(end)
        self.styling.opacity.set(start, end, lambda t: end_val * easing((t-start)/(end-start)))

    def write(self, start=0, end=1, max_stroke_width=1, change_existence=True, easing=easings.linear, stroke_easing=easings.there_and_back):
        """This object will get to the original fill opacity at the end starting from none
        The stroke width will increase and decrease
        The show argument before start will be set to none if change_existence, this is a creating function"""
        # This creates the object, we can thus not see it before this command
        if change_existence:
            self.show.set_onward(0, False)
            self.show.set_onward(start, True)

        end_val = self.styling.fill_opacity.at_time(end)
        self.styling.fill_opacity.set(start, end, lambda t: end_val * easing((t-start)/(end-start)))

        # Increase and decrease the stroke_width for a look at me affect
        sw = self.styling.stroke_width.at_time(end)
        # self.styling.stroke_width.set(start, end, lambda t: max_stroke_width*4*(0.5-abs(linear()(t)-.5))**2 + linear()(t) * sw)
        # self.styling.stroke_width.set(start, end, lambda t: max_stroke_width*4*(0.5-abs(easing((t-start)/(end-start))-.5))**2 + easing((t-start)/(end-start)) * sw)
        s, e = start, end
        self.styling.stroke_width.set(s, e, lambda t: max_stroke_width * stroke_easing((t-s)/(e-s)) + easing((t-s)/(e-s)) * sw)

    def create(self, start=0, end=1, change_existence=True, easing=easings.smooth):
        """Creation argument, draw the path, then do a fade in
        NOTE: add the returned object to the canvas"""
        # NOTE: TODO: creation, simplify and rewrite
        # This creates the object, we can thus not see it before this command
        if change_existence:
            self.show.set_onward(0, False)
            self.show.set_onward(end, True)

        p = morphing.Path(self.path(start))

        def sample(path, t):
            """Sample a path a time 0<=t<=1 based on amount of curves"""
            tot_secs = len(path)
            time = t * tot_secs
            segs = []
            idx = 0
            while time >= 1:
                segs.append(path[idx])
                time -= 1
                idx += 1
            segs.append(path[idx].split(time)[0])
            return morphing.Path(*segs)

        def sample2(path, t):
            """Sample a path a time 0<=t<=1 based on length of the curve"""
            tot_length = path.length()
            length_to_draw = t * tot_length
            segs = []
            idx = 0
            while length_to_draw > 0:
                s = path[idx]
                l = s.length()
                if l <= length_to_draw:
                    segs.append(s)
                    length_to_draw -= l
                else:
                    # Split the segment at the length
                    segs.append(s.split(length_to_draw / l)[0])
                    length_to_draw = 0
                idx += 1
            return morphing.Path(*segs)

        assert callable(easing)
        def f(t): return easing((t-start)/(end-start))

        res = Path('')
        res.d.set(start, end, lambda t: sample2(p, f(t)).d())
        res.styling = self.styling
        res.styling.fill_opacity.set_onward(0, 0)
        res.styling.stroke.set_onward(0, 'green')
        res.styling.stroke_width.set_onward(0, 10)
        if change_existence:
            res.show.set_onward(0, False)
            res.show.set_onward(start, True)
        print('NOTE: Creating an object is under development')
        return res


class Polygon(VObject):
    def __init__(self, *vertices, z=0, creation=0, **styling_kwargs):
        super().__init__(creation=creation, z=z)
        # The vertices are automatically closed
        self.vertices = [attributes.Coor(creation, v) for v in vertices]
        # The styling arguments
        self.styling = style.Styling(styling_kwargs, creation=creation, fill_opacity=0.7, stroke='#fff', stroke_width=2)

    @property
    def last_change(self):
        return max(max(v.last_change for v in self.vertices), self.z.last_change, self.styling.last_change, self.show.last_change)

    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        if end_time == None:  # We do not animate this movement
            for v in self.vertices:
                v.add_onward(start_time, (dx, dy))
        else:
            raise NotImplementedError('Animation for shifting a Polygon is not yet implemented')
        return self

    def path(self, time):
        vert = [v.at_time(time) for v in self.vertices]
        path = f'M {vert[0][0]},{vert[0][1]}'
        for x, y in vert:
            path += f' L {x},{y}'
        path += ' Z'
        return path

    def to_svg(self, time):
        string = '<polygon'
        # We need to define the points of the Polygon
        points = [v.at_time(time) for v in self.vertices]
        string += f" points='{' '.join(f'{x},{y}' for x, y in points)}'"
        # Add the style
        string += self.styling.svg_style(time)
        # End the object
        string += ' />'
        return string


class FixedVertexPolygon(VObject):
    """The vertices are fixed in time and cannot change"""
    def __init__(self, *vertices, z=0, creation=0, **styling_kwargs):
        super().__init__(creation=creation, z=z)
        self.creation = creation
        # The vertices are automatically closed
        self.vertices = vertices
        self.vertex_string = f" points='{' '.join(f'{x},{y}' for x, y in vertices)}'"
        # The styling arguments
        self.styling = style.Styling(styling_kwargs, creation=creation, fill_opacity=0.7, stroke='#fff', stroke_width=2)

    @property
    def last_change(self):
        return max(self.creation, self.styling.last_change, self.z.last_change, self.show.last_change)

    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        raise NotImplementedError('Shifting a Fixed Polygon is not implemented')

    def path(self, time):
        raise NotImplementedError('Retrieving the path of a Fixed Polygon is not implemented')

    def to_svg(self, time):
        string = '<polygon'
        # We need to define the points of the Polygon
        string += self.vertex_string
        # Add the style
        string += self.styling.svg_style(time)
        # End the object
        string += ' />'
        return string


class Circle(VObject):
    def __init__(self, r=100, cx=500, cy=500, z=0, creation=0, **styling_kwargs):
        super().__init__(creation=creation, z=z)
        # Load the values to initalize the shape
        self.c = attributes.Coor(creation, (cx, cy))
        self.r = attributes.Real(creation, r)
        # The styling arguments
        self.styling = style.Styling(styling_kwargs, creation=creation, fill='#e07a5f', fill_opacity=0.7, stroke='#e07a5f', stroke_width=5)

    @property
    def last_change(self):
        return max(self.c.last_change, self.r.last_change, self.styling.last_change, self.z.last_change, self.show.last_change)

    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        if end_time == None:  # We do not animate this movement
            self.c.add_onward(start_time, (dx, dy))
        else:
            assert callable(easing)
            s, e = start_time, end_time
            self.c.add_onward(s, lambda t: (dx * easing((t-s) / (e-s)), dy * easing((t-s) / (e-s))), last_change=e)
        return self

    def path(self, time):
        # SEE: https://github.com/alexk111/SVG-Morpheus/blob/master/source/js/svg-morpheus.js line 1131
        r = self.r.at_time(time)
        cx, cy = self.c.at_time(time)
        path = f'M{cx-r},{cy}a{r},{r} 0 1,0 {r*2},0a{r},{r} 0 1,0 -{r*2},0z';
        return path

    def to_svg(self, time):
        string = '<circle'
        # We need to define the points of the Polygon
        string += f" cx='{self.c.at_time(time)[0]}'"
        string += f" cy='{self.c.at_time(time)[1]}'"
        string += f" r='{self.r.at_time(time)}'"
        # Add the style
        string += self.styling.svg_style(time)
        # End the object
        string += ' />'
        return string


class Dot(Circle):
    """Returns a point (a small circle with different default styling)"""
    # NOTE: TODO convert Dot into Circle with different default styling arguments
    def __init__(self, r=6, cx=500, cy=500, z=0, creation=0, **styling_kwargs):
        new_style = {'fill': '#83C167', 'fill_opacity': 0.7, 'stroke_width': 0} | styling_kwargs
        super().__init__(r=r, cx=cx, cy=cy, z=z, creation=creation, **new_style)

class Rectangle(VObject):
    def __init__(self, width, height, x=500, y=500, rx=0, ry=0, creation=0, z=0, **styling_kwargs):
        super().__init__(creation=creation, z=z)
        # Load the values to initalize the shape
        self.x = attributes.Real(creation, x)
        self.y = attributes.Real(creation, y)
        self.width = attributes.Real(creation, width)
        self.height = attributes.Real(creation, height)
        # Define the rounding of the corners
        self.rx = attributes.Real(creation, rx)
        self.ry = attributes.Real(creation, ry)
        # The styling arguments
        self.styling = style.Styling(styling_kwargs, creation=creation, fill_opacity=0.7, stroke='#fff', stroke_width=3)

    @property
    def last_change(self):
        return max(self.x.last_change, self.y.last_change, self.width.last_change,
                   self.height.last_change, self.rx.last_change, self.ry.last_change,
                   self.styling.last_change, self.z.last_change, self.show.last_change)

    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        if end_time == None:  # We do not animate this movement
            self.x.add_onward(start_time, dx)
            self.y.add_onward(start_time, dy)
        else:
            assert callable(easing)
            s, e = start_time, end_time
            self.x.add_onward(s, lambda t: dx * easing((t-s) / (e-s)), last_change=e)
            self.y.add_onward(s, lambda t: dy * easing((t-s) / (e-s)), last_change=e)
        return self

    def path(self, time):
        x, y = self.x.at_time(time), self.y.at_time(time)
        w, h = self.width.at_time(time), self.height.at_time(time)
        rx, ry = self.rx.at_time(time), self.ry.at_time(time)
        # SEE: https://github.com/alexk111/SVG-Morpheus/blob/master/source/js/svg-morpheus.js lines 1147-1159
        if rx == 0 and ry == 0:
            path = f'M{x},{y}l{w},0l0,{h}l-{w},0z';
        else:
            # raise NotImplementedError('See below')
            path = f'M{x+rx},{y} l {w-rx*2},0 ' \
                   f'a {rx},{ry} 0 0,1 {rx},{ry} l 0,{h-ry*2} ' \
                   f'a {rx},{ry} 0 0,1 -{rx},{ry} l {rx*2-w},0 ' \
                   f'a {rx},{ry} 0 0,1 -{rx},-{ry} l 0,{ry*2-h} ' \
                   f'a {rx},{ry} 0 0,1 {rx},-{ry} z'
        return path

    def to_svg(self, time):
        string = '<rect'
        # We need to define the points of the Polygon
        string += f" x='{self.x.at_time(time)}'"
        string += f" y='{self.y.at_time(time)}'"
        string += f" width='{self.width.at_time(time)}'"
        string += f" height='{self.height.at_time(time)}'"
        string += f" rx='{self.rx.at_time(time)}'"
        string += f" ry='{self.ry.at_time(time)}'"
        # Add the style
        string += self.styling.svg_style(time)
        # End the object
        string += ' />'
        return string


class EquilateralTriangle(Polygon):
    # NOTE: TODO: ADD ROTATION OPTION
    def __init__(self, side_length, angle=0, cx=500, cy=500, creation=0, z=0, **styling_kwargs):
        """The center is the center of mass at 1/3 the height of the triangle"""
        s = side_length
        h = math.sqrt(3)/2 * s
        vertices = [(cx-s/2, cy+h/3), (cx+s/2, cy+h/3), (cx, cy-2*h/3)]
        def rotate(origin, point, angle):
            """Rotate a point counterclockwise by a given angle around a given origin."""
            ox, oy = origin
            px, py = point
            qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
            qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
            return qx, qy
        midpoint = (sum(v[0] for v in vertices)/3, sum(v[1] for v in vertices)/3)
        vertices = [rotate(midpoint, v, angle*math.pi/180) for v in vertices]
        super().__init__(*vertices, creation=creation, z=z, **styling_kwargs)


class Line(VObject):
    def __init__(self, x1=0, y1=0, x2=100, y2=100, creation=0, z=0, **styling_kwargs):
        super().__init__(creation=creation, z=z)
        # Load the values to initalize the shape
        self.p1 = attributes.Coor(creation, (x1, y1))
        self.p2 = attributes.Coor(creation, (x2, y2))
        # The styling arguments
        self.styling = style.Styling(styling_kwargs, creation=creation, stroke='#fff', stroke_width=3)

    @property
    def last_change(self):
        return max(self.p1.last_change, self.p2.last_change, self.styling.last_change, self.z.last_change, self.show.last_change)

    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        if end_time == None:  # We do not animate this movement
            self.p1.add_onward(start_time, (dx, dy))
            self.p2.add_onward(start_time, (dx, dy))
        else:
            assert callable(easing)
            s, e = start_time, end_time
            self.p1.add_onward(s, lambda t: (dx * easing((t-s)/(e-s)), dy * easing((t-s)/(e-s))), last_change=e)
            self.p2.add_onward(s, lambda t: (dx * easing((t-s)/(e-s)), dy * easing((t-s)/(e-s))), last_change=e)
        return self

    def path(self, time):
        x1, y1 = self.p1.at_time(time)
        x2, y2 = self.p2.at_time(time)
        return f'M{x1},{y1}L{x2},{y2}Z'

    def to_svg(self, time):
        string = '<line'
        # We need to define the points of the Polygon
        string += f" x1='{self.p1.at_time(time)[0]}'"
        string += f" y1='{self.p1.at_time(time)[1]}'"
        string += f" x2='{self.p2.at_time(time)[0]}'"
        string += f" y2='{self.p2.at_time(time)[1]}'"
        # Add the style
        string += self.styling.svg_style(time)
        # End the object
        string += ' />'
        return string


class Lines(VObject):
    def __init__(self, *vertices, creation=0, z=0, **styling_kwargs):
        super().__init__(creation=creation, z=z)
        self.vertices = [attributes.Coor(creation, v) for v in vertices]
        # The styling arguments
        self.styling = style.Styling(styling_kwargs, creation=creation, stroke='#fff', stroke_width=3)

    @property
    def last_change(self):
        return max(max(v.last_change for v in self.vertices), self.styling.last_change, self.z.last_change, self.show.last_change)

    def to_svg(self, time):
        string = '<polyline'
        # We need to define the points of the Polygon
        points = [v.at_time(time) for v in self.vertices]
        string += f" points='{' '.join(f'{x},{y}' for x, y in points)}'"
        # Add the style
        string += self.styling.svg_style(time)
        # End the object
        string += ' />'
        return string


class Trace(VObject):
    """Folows the point every dt and converts to lines (every dt time)"""
    def __init__(self, point, start=0, end=None, dt=1/60, z=0, **styling_kwargs):
        super().__init__(creation=start, z=z)
        # Load the values to initalize the shape
        self.start = start
        self.end = end
        self.dt = dt
        self.p = point
        # The styling arguments
        self.styling = style.Styling(styling_kwargs, creation=start, stroke='#fff', stroke_width=3)
        self.vert_cache = []
        self.vert_string_cache = ''
        self.vert_string_positions = [-1]

    @property
    def last_change(self):
        return max(self.p.last_change, self.styling.last_change, self.z.last_change, self.show.last_change)

    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        raise NotImplementedError('Shifting a Trace object is not implemented.')

    def path(self, time):
        raise NotImplementedError('Retrieving the path of a Trace object is not implemented.')
        # x1, y1 = self.p1.at_time(time)
        # x2, y2 = self.p2.at_time(time)
        # return f'M{x1},{y1}L{x2},{y2}Z'

    def vertices(self, time):
        """Get the vertices of this path following the point
        i.e. this is equal to the positions of this point spaced dt apart"""
        t = self.start
        end = self.end
        if end == None:  # Indefinite trace, forever
            end = time
        end = min(end, time)
        steps = int((end-t)/self.dt)
        # Add from cache
        vert = self.vert_cache[:steps]
        l = len(vert)
        t += l * self.dt
        done_something = False
        for _ in range(steps - l):
            tmp = self.p.at_time(t)
            vert.append(tmp)
            t += self.dt
            done_something = True
        # Update cache, but only if new values are found
        if done_something:
            self.vert_cache = vert
            for x, y in vert[l:]:
                tmp = f'{x},{y}'
                self.vert_string_cache += tmp + ' '
                self.vert_string_positions.append(self.vert_string_positions[-1] + len(tmp)+1)
        return vert

    def to_svg(self, time):
        string = '<polyline'
        self.vertices(time)  # Set cache values
        end = min(self.end, time) if self.end != None else time
        steps = int((end - self.start)/self.dt)
        if steps == 0:
            return ''
        point_str = self.vert_string_cache[:self.vert_string_positions[steps]]
        current_pos = self.p.at_time(time)
        string += f" points='{point_str} {current_pos[0]},{current_pos[1]}'"
        # Add the style
        string += self.styling.svg_style(time)
        # End the object
        string += ' />'
        return string

    def to_polygon(self, time, fixed=True):
        """We convert this trace into a polygon"""
        if fixed:  # For render speed
            return FixedVertexPolygon(*self.vertices(time), creation=time, z=self.z.at_time(time), **self.styling.kwargs())
        else:
            return Polygon(*self.vertices(time), creation=time, z=self.z.at_time(time), **self.styling.kwargs())

class VCollection:
    """This is the motherclass for all vector objects
    Since an image is not a vector object this will not be a child
    (thus the fadein function and fadeout function will not be the
     same as for vector objects)
    """
    def __init__(self, *objects, creation=0, z=0):
        self.objects = list(objects)
        # If we need to show the entire group
        self.show = attributes.Real(creation, True)
        # Set the draw order
        self.z = attributes.Real(creation, z)

    @property
    def last_change(self):
        return max(max(obj.last_change for obj in self.objects), self.z.last_change, self.show.last_change)

    def __iter__(self):
        return iter(self.objects)

    def to_svg(self, time):
        # We encapsulate all objects into a g object
        string = '<g'
        string += '>\n'
        # Sorting based on the z property/draw order
        lst = []
        for obj in self.objects:
            if obj.show.at_time(time):
                if not hasattr(obj, 'z'):
                    lst.append((0, obj))
                else:
                    lst.append((obj.z.at_time(time), obj))

        # Add the objects
        for _, obj in sorted(lst, key=lambda x: x[0]):
            string += obj.to_svg(time)
            string += '\n'
        string += '</g>'

        return string

    def bbox(self, time, start_idx=0, end_idx=None):
        """Get the bounding box (x, y, width, height of the collection),
        no end_idx means all objects starting from start_idx"""
        if end_idx == None:
            end_idx = len(self.objects)
        obj_box = self.objects[start_idx].bbox(time)
        xmin, ymin = obj_box[0], obj_box[1]
        xmax, ymax = obj_box[2]+xmin, obj_box[3]+ymin
        for obj in self.objects[start_idx+1:end_idx]:
            obj_box = obj.bbox(time)
            xmin = min(xmin, obj_box[0])
            ymin = min(ymin, obj_box[1])
            xmax = max(xmax, obj_box[2]+obj_box[0])
            ymax = max(ymax, obj_box[3]+obj_box[1])
        return (xmin, ymin, xmax-xmin, ymax-ymin)

    def brect(self, time, start_idx=0, end_idx=None, rx=0, ry=0, dpos=5, follow=True):
        """Gets the bounding rectangle (or surrounding rectangle)
        Adjusted with dpos outward, no ending index means all objects"""
        if end_idx == None:
            end_idx = len(self.objects)

        if not follow:  # Get the bounding rectangle once
            x, y, width, height = self.bbox(time, start_idx, end_idx)
            return Rectangle(
                width+2*dpos, height+2*dpos, x-dpos, y-dpos, rx=rx, ry=ry,
                fill_opacity=0, stroke_opacity=1, stroke='#ff0', stroke_width=2
            )
        else:
            rect = Rectangle(width=0, height=0, rx=rx, ry=ry,
                             fill_opacity=0, stroke_opacity=1, stroke='#ff0', stroke_width=2)
            rect.x.set_onward(time, lambda t: self.bbox(t, start_idx, end_idx)[0]-dpos)
            rect.y.set_onward(time, lambda t: self.bbox(t, start_idx, end_idx)[1]-dpos)
            rect.width.set_onward(time, lambda t: self.bbox(t, start_idx, end_idx)[2]+2*dpos)
            rect.height.set_onward(time, lambda t: self.bbox(t, start_idx, end_idx)[3]+2*dpos)
            return rect

    def center_to_pos(self, posx=500, posy=500, start_time=0, end_time=None, easing=easings.smooth):
        """Shifts the center to pos, and animates this from start_time to end_time
        Shifting is done with respect to the position at start_time."""
        xmin, ymin, width, height = self.bbox(start_time)
        dx, dy = posx-(xmin+width/2), posy-(ymin+height/2)
        self.shift(dx=dx, dy=dy, start_time=start_time, end_time=end_time, easing=easing)

    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        """Shift all objects"""
        for obj in self.objects:
            obj.shift(dx=dx, dy=dy, start_time=start_time, end_time=end_time, easing=easing)

    def fadein(self, **kwargs):
        """Fade in the group of VObjects"""
        for obj in self.objects:
            obj.fadein(**kwargs)

    def write(self, start, end, processing=10, max_stroke_width=1/2, change_existence=True):
        """Write the group of VObjects with a sliding window at a time"""
        # Define the seconds per char
        spc = (end-start) / (len(self.objects) + processing)
        # Write all the objects in the sliding window
        for num, obj in enumerate(self.objects):
            obj.write(start=start+spc*num, end=start+spc*(num+processing+1), max_stroke_width=max_stroke_width, change_existence=change_existence)


class MorphObject(VCollection):
    """This morphes one object into the other with a starting time and ending time, this, itself, is an object
    this also needs to be added to the canvas."""
    def __init__(self, morph_from, morph_to, start=0, end=1, z=0, easing=easings.smooth, change_existence=True):
        # Both morph_from and morph_to are converted to collections
        if isinstance(morph_from, VObject):
            morph_from = VCollection(morph_from)
        if isinstance(morph_to, VObject):
            morph_to = VCollection(morph_to)
        assert isinstance(morph_from, VCollection)
        assert isinstance(morph_to, VCollection)

        # This is a converting object, the starting object is not visible during or after the transformation
        # and the ending object is not visible before or during the conversion. If change_existency is True
        # we enforce these existences on the objects.
        if change_existence:
            for obj in morph_from:
                # This is a destruction object for morph_from
                obj.show.set_onward(start, False)
            for obj in morph_to:
                # This is a creation object for morph_to
                obj.show.set_onward(0, False)
                obj.show.set_onward(end, True)

        # Get the starting paths of all objects
        paths_from = [(morphing.Path(obj.path(start)), obj.styling) for obj in morph_from]
        paths_to = [(morphing.Path(obj.path(end)), obj.styling) for obj in morph_to]
        # Get the objects in theoretical paths which be need to convert/morph
        obj_from = morphing.Paths(*paths_from)
        obj_to = morphing.Paths(*paths_to)

        # NOTE: TODO: add transforms, i.e. adjusted paths and rotations
        mapping = obj_from.morph(obj_to, start_time=start, end_time=end, easing=easings.smooth)

        # We now initialize the parent with objects we get from the mapping
        # First we need to convert the mapping into an array of displayable objects.
        objects = []
        for path_specifyer, styling_from, styling_to in mapping:
            new = Path('', x=0, y=0, creation=start, z=z)
            # We need to destroy this morphing object after the animation ended
            new.show.set_onward(end, False)
            # Set the path_specifyer of the path object to equal the mapping
            # Note that the returned specifyer must be remapped from 0<=t<=1 to start<=t<=end
            def f(specifyer):
                return lambda t: specifyer((t-start)/(end-start))
            new.d.set(start, end, f(path_specifyer))
            # We must now set the styling arguments between start and end, morphing styling_from to styling_to
            # But we cannot alter any object from either of the stylings as that would change stylings in other
            # objects as well.
            new.styling = styling_from.interpolate(styling_to, start, end, easing=easing)
            # Add as an object to the collection (of subpaths)
            objects.append(new)

        # Initialize a collection of these objects
        super().__init__(*objects, creation=start, z=z)
        # Discard the entire collection
        self.show.set_onward(end, False)

# :::MARKER: Optimisation/rewrite of code ended here
# NOTE: TODO: add z-coordinates / draw order, argument
# NOTE: TODO: remember, z coordinate as argument + in last_change

class TexObject(VCollection):
    def __init__(self, to_render, x=0, y=0, width=None, creation=0, **styles):
        # NOTE: only one of (width, height), scale can be defined
        # NOTE: make a wildcard VObject that's just an svg for reading other svgs, etc
        # Get the characters and parse them as VObjects
        tex_dir = f'{save_directory}/tex'
        self.char_viewbox, chars = get_characters(tex_dir, to_render, 'latex', '')
        st = {'stroke_width': 0, 'fill': '#fff'} | styles
        # st = style.Styling(styles, fill='#fff', fill_opacity=1, stroke_width=1, stroke='#fff').kwargs()
        chars = [from_svg(char, **st) for char in chars]

        # Init the collection of VObjects
        super().__init__(*chars, creation=creation)

        # Initialize the position and scale/width of the group viewbox
        self.x = attributes.Real(creation, x)
        self.y = attributes.Real(creation, y)

        # Get the bounding box and reset the position
        xmin, ymin, _, _ = self.bbox(creation)
        for obj in self.objects:
            obj.styling.dx.add_onward(creation, lambda t: self.x.at_time(t) - xmin)
            obj.styling.dy.add_onward(creation, lambda t: self.y.at_time(t) - ymin)


class Path(VObject):
    """Similar to a svg path"""
    def __init__(self, path, x=0, y=0, creation=0, z=0, **styling_kwargs):
        super().__init__(creation=creation, z=z)
        # Load the values to initalize the shape
        self.d = attributes.String(creation, path)
        # The styling arguments
        self.styling = style.Styling(styling_kwargs, creation=creation, stroke='#fff')
        # Add the position as the transform translate argument after the scale
        self.styling.dx.add_onward(creation, lambda t: x * self.styling.scale_x.at_time(t))
        self.styling.dy.add_onward(creation, lambda t: y * self.styling.scale_y.at_time(t))

    @property
    def last_change(self):
        return max(self.d.last_change, self.styling.last_change, self.z.last_change, self.show.last_change)

    def shift(self, dx=0, dy=0, start_time=0, end_time=None, easing=easings.smooth):
        """Shift the object dx and dy (by changing the transformation styling)"""
        if end_time == None:  # We do not animate this movement
            self.styling.dx.add_onward(start_time, dx)
            self.styling.dy.add_onward(start_time, dy)
        else:
            raise NotImplementedError('Animation for shift not implemented, get smoothing function here')
        return self

    def path(self, time):
        return self.d.at_time(time)

    def to_svg(self, time):
        string = '<path'
        # We need to define the points of the Polygon
        string += f" d='{self.d.at_time(time)}'"
        # Add the style
        string += self.styling.svg_style(time)
        # End the object
        string += ' />'
        return string


def from_svg(specifyer, **styles):
    """Go from bs4 to VObject"""
    if specifyer.name == 'path':
        path = specifyer['d']
        del specifyer['d']
        specifyer['x'] = float(specifyer['x'])
        specifyer['y'] = float(specifyer['y'])
        return Path(path, **(styles | specifyer.attrs))
    elif specifyer.name == 'rect':
        width = float(specifyer['width'])
        height = float(specifyer['height'])
        del specifyer['width']
        del specifyer['height']
        specifyer['x'] = float(specifyer['x'])
        specifyer['y'] = float(specifyer['y'])
        return Path(Rectangle(width, height, x=0, y=0).path(0), **(styles | specifyer.attrs))
        # print(Rectangle(width, height, x=0, y=0).path(0))
        return Rectangle(width, height, **(styles | specifyer.attrs))
    else:
        raise NotImplementedError(f'Type "{specifyer.name}" has no from_svg implemented')


# NOTE: TODO: add speed up buttons and FPS info, etc
# NOTE: TODO: end_time = 0 signifies a single picture
# NOTE: TODO: make a split-tex object that returns multiple vcollections and is thus easier for the MOVING FRAME BOX EXAMPLE
# NOTE: TODO: Update attributes functionalities
# NOTE: TODO: Write to png and mp4 and separate svgs
# NOTE: TODO: Make text object
# NOTE: TODO: Make image object
# NOTE: TODO: Add arc in morphing file
# NOTE: TODO: Restructure files
# NOTE: TODO: Morph with rotate
# NOTE: TODO: Implement create, draw path elements and apply filling.
# NOTE: TODO: add ellipse
# NOTE: TODO: check on other svg objects
# NOTE: add documentation on styling arguments SEE: 'PRESENTATION ATTRIBUTES': https://developer.mozilla.org/en-US/docs/Web/SVG/Element/polygon
# NOTE: TODO: ADD MORE STYLES / PRESENTATION ATTRIBUTES SEE: https://developer.mozilla.org/en-US/docs/Web/SVG/Element/polygon
# NOTE: TODO: add more checks to the code
# NOTE: TODO: add keys like 1,2,3,4,5,6,7,8,9 that signify a time in the view (as percentage or so)
# NOTE: TODO: enable hot-reload on saving the file in the window
# NOTE: TODO: let the user specify easing functions for move_to, shift, etc...
# NOTE: TODO: enable x,y or coordinate setting (also as attribute)
# NOTE: TODO: Correct bounding boxes for transformations (including rotation) -> in morphing
# NOTE: TODO: Coordinates to complex values?
# NOTE: TODO: brect is very slow with latex and follow=True, because of bbox and all its rotations? in that case maybe temporarily store the bbox until a change is made to the object? Or can other optimisations be made in e.g. calculating the box or defining the functions that go with it (maybe on function returning two functions or so is faster).
# NOTE: TODO: VCollection make assertion that it is an iterable of VObject
# NOTE: TODO: Enable preprocessing so that fps is exactly as specified -> already call frame contructions
# NOTE: TODO: EXAMPLE MOVING FRAME BOX with scale_x and scale_y and center_to_pos causes rectangles in tex to move somewhere else
# NOTE: TODO: Add copy functions
# NOTE: TODO: fix the attributes
# NOTE: TODO: write svgs to folder with pause metadata
# NOTE: TODO: pause with slide numbering
# NOTE: TODO: start and end consistently
# NOTE: TODO: round values pushed to svg for compression of data
# NOTE: TODO: compress the folder of svgs created and unfold in run-time
# NOTE: TODO: when using pauses and "slides" get function to export first slide of animation to a more normal-human-readable format
# NOTE: TODO: make a minimal SVGViewer without the toolbars etc., just the keyboard functions.



"""
MIT License

Copyright (c) 2018 3Blue1Brown LLC
Copyright (c) 2021, the Manim Community Developers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import os
import hashlib
from copy import copy
from bs4 import BeautifulSoup


def hash_string(string: str) -> str:
    # Truncating at 16 bytes for cleanliness
    hasher = hashlib.sha256(string.encode())
    return hasher.hexdigest()[:16]

def tex_content_to_svg_file(
    dir_name: str, content: str, compiler: str, preamble: str,
) -> str:

    full_tex = "\n\n".join((
        "\\documentclass[preview]{standalone}",
        "\\usepackage{amsmath}",
        "\\usepackage{amsfonts}",
        preamble,
        "\\begin{document}",
        content,
        "\\end{document}"
    )) + "\n"

    svg_file = os.path.join(
        dir_name, hash_string(full_tex) + ".svg"
    )
    if not os.path.exists(svg_file):
        # If svg doesn't exist, create it
        # with display_during_execution("Writing " + short_tex):
        os.makedirs(dir_name, exist_ok=True)
        create_tex_svg(full_tex, svg_file, compiler)
    return svg_file


def create_tex_svg(full_tex: str, svg_file: str, compiler: str) -> None:
    if compiler == "latex":
        program = "latex"
        dvi_ext = ".dvi"
    elif compiler == "xelatex":
        program = "xelatex -no-pdf"
        dvi_ext = ".xdv"
    else:
        raise NotImplementedError(
            f"Compiler '{compiler}' is not implemented"
        )

    # Write tex file
    root, _ = os.path.splitext(svg_file)
    with open(root + ".tex", "w", encoding="utf-8") as tex_file:
        tex_file.write(full_tex)

    # tex to dvi
    if os.system(" ".join((
        program,
        "-interaction=batchmode",
        "-halt-on-error",
        f"-output-directory=\"{os.path.dirname(svg_file)}\"",
        f"\"{root}.tex\"",
        ">",
        os.devnull
    ))):
        log.error(
            "LaTeX Error!  Not a worry, it happens to the best of us."
        )
        error_str = ""
        with open(root + ".log", "r", encoding="utf-8") as log_file:
            error_match_obj = re.search(r"(?<=\n! ).*\n.*\n", log_file.read())
            if error_match_obj:
                error_str = error_match_obj.group()
                log.debug(
                    f"The error could be:\n`{error_str}`",
                )
        raise LatexError(error_str)

    # dvi to svg
    os.system(" ".join((
        "dvisvgm",
        f"\"{root}{dvi_ext}\"",
        "-n",
        "-v",
        "0",
        "-o",
        f"\"{svg_file}\"",
        ">",
        os.devnull
    )))

    # Cleanup superfluous documents
    for ext in (".tex", dvi_ext, ".log", ".aux"):
        try:
            os.remove(root + ext)
        except FileNotFoundError:
            pass

def get_characters(tex_dir, to_render, compiler, preamble):
    # Write the tex content to an svg file
    filename = tex_content_to_svg_file(tex_dir, content=to_render, compiler='latex', preamble='')

    with open(filename, 'r') as f:
        svg_contents = BeautifulSoup(f.read(), features='xml').svg
        surrounding_box = [float(i) for i in svg_contents['viewBox'].split(' ')]

        chars = [svg_char for svg_char in svg_contents.g if svg_char != '\n']

        # Some chars use definitions, we add the value of the 'use' to the definition
        uses = [(ref_id[1:], c, chard_idx) for chard_idx, c in enumerate(chars) if (ref_id := c.get('xlink:href', None)) != None]

        defs = {i.get('id', None): i for i in svg_contents.defs if i != '\n'}
        for def_id, use, char_idx in uses:
            # Find the corresponding def to the use
            d = copy(defs[def_id])
            # d = svg_contents.select(f'#{def_id}', )[0]
            del use['xlink:href']
            del d['id']
            for name, val in use.attrs.items():
                if d.get(name, None) != None:
                    # Add the two values
                    d[name] = str(float(d[name]) + float(val))
                else:
                    d[name] = val

            chars[char_idx] = d

    return surrounding_box, chars


if __name__ == '__main__':
    tex_content_to_svg_file('svgs/scene1/tex', '$$x^2$$', compiler='latex', preamble='')




import math
import vectormation.colors as colors
import vectormation.easings as easings

def sigmoid(x):
  return 1 / (1 + math.exp(-x))


def heaviside(time, if_equal=1):
    """Returns a heaviside/step function being 0 if x < time and 1 if x > time"""
    return lambda x: 1 if x > time else 0 if x < time else if_equal


class Real:
    """The real attribute is a real-valued attribute (changing) through time"""
    def __init__(self, creation, start_val=0):
        if isinstance(start_val, Real):
            self.set_to(start_val)
        else:
            # Define the function
            self.time_func = lambda t: heaviside(creation)(t) * start_val
            # Define the last change time of this attribute (after that it's constant)
            self.last_change = creation

    def set(self, start, end, func_inner, lincl=True, rincl=True, stay=False):
        """Set this attribute to func_inner(x) on [start, end]
        lincl, rincl are if the brackets are [/] or (/)"""
        func_outer = self.time_func
        def new_func(t):
            if t < start: return func_outer(t)
            elif t > end:
                if stay: return new_func(end)
                else: return func_outer(t)
            elif not lincl and t == start: return func_outer(t)
            elif not rincl and t == end: return func_outer(t)
            else: return func_inner(t)
        self.time_func = new_func
        self.last_change = max(self.last_change, end)

    def add(self, start, end, func_inner, lincl=True, rincl=True, stay=False):
        """Set this attribute to func_inner(x)+previous on [start, end]
        lincl, rincl are if the brackets are [/] or (/)"""
        func_outer = self.time_func
        def new_func(t):
            if t < start: return func_outer(t)
            elif t > end:
                if stay: return new_func(end)
                else: return func_outer(t)
            elif not lincl and t == start: return func_outer(t)
            elif not rincl and t == end: return func_outer(t)
            else: return float(func_outer(t)) + func_inner(t)
        self.time_func = new_func
        self.last_change = max(self.last_change, end)

    def set_onward(self, start, value, lincl=True):
        """Set this attribute to value on [start, infty]
        lincl are if the bracket is [/] or (/)"""
        func_left = self.time_func
        def new_func(t):
            if t < start: return func_left(t)
            elif not lincl and t == start: return func_left(t)
            else:
                if callable(value): return value(t)
                else: return value
        self.time_func = new_func
        self.last_change = max(self.last_change, start)

    def add_onward(self, start, func, lincl=True, last_change=None):
        """Set this attribute to value + func(val) on [start, infty]
        lincl are if the bracket is [/] or (/)"""
        func_left = self.time_func
        def new_func(t):
            if t < start: return func_outer(t)
            elif not lincl and t == start: return func_left(t)
            else:
                if callable(func): return float(func_left(t)) + func(t)
                else: return float(func_left(t)) + func
        self.time_func = new_func
        self.last_change = max(self.last_change, start)
        if last_change != None:
            self.last_change = max(self.last_change, last_change)


    def set_at(self, time, value):
        """Set this attribute to value on time
        lincl are if the bracket is [/] or (/)"""
        func_otherwise = self.time_func
        def new_func(t):
            if t == time: return value
            else: return func_otherwise(t)
        self.time_func = new_func
        self.last_change = max(self.last_change, time)

    def add_at(self, time, value, default=0):
        """Set this attribute to the value + the value before on time
        If the value is none, set to default.
        lincl are if the bracket is [/] or (/)"""
        old = float(self.time_func(time))
        if old == None:
            old = default
        func_otherwise = self.time_func
        def new_func(t):
            if t == time: return old + value
            else: return func_otherwise(t)
        self.time_func = new_func
        self.last_change = max(self.last_change, time)

    def at_time(self, time):
        return self.time_func(time)

    def set_to(self, other):
        """Set the value of self to always be the value of other"""
        # The value needs to update as the time_function updates
        self.time_func = lambda t: other.time_func(t)
        self.last_change = other.last_change

    def move_to(self, start_time, end_time, end_val, stay=True, easing=easings.smooth):
        """Animate the object going from start_pos to end_pos, exponentially"""
        start_val = self.time_func(start_time)
        diff = start_val - end_val
        s, e = start_time, end_time
        self.set(s, e, lambda t: diff * (1-easing((t-s)/(e-s))) + end_val, stay=stay)
        self.last_change = max(self.last_change, end_time)

    def interpolate(self, other, start, end, easing=easings.linear):
        """Interpolate this value from start to end with easing to the other value
        Return a lambda function (or this same attribute) for these times"""
        assert self.__class__ == Real  # Both classes need to be a real object
        assert other.__class__ == Real
        start_val = self.time_func(start)
        end_val = other.time_func(end)
        # We need to get from start_val to end_val with easing
        s, e = start, end

        new = Real(0)
        new.time_func = lambda t: start_val + (end_val-start_val) * easing((t-s)/(e-s))
        new.last_change = end
        return new


class Natural(Real):
    """A natural number"""
    def at_time(self, time, rounding_func=round):
        """Get the value at some time"""
        return rounding_func(self.time_func(time))


class Tup(Real):
    """Arbitrary length real attribute"""
    def __init__(self, creation, start_val):
        if isinstance(start_val, Tup):
            self.set_to(start_val)
        else:
            # Define the function
            self.time_func = lambda t: tuple(
                    [heaviside(creation)(t) * val for val in start_val]
                )
            # Define the last change time of this attribute (after that it's constant)
            self.last_change = creation

    def add(self, start, end, value, lincl=True, rincl=True, stay=False):
        """Set this attribute to value+previous on [start, end]
        lincl, rincl are if the brackets are [/] or (/)"""
        # NOTE: TODO: complete rewrite maybe with an Attribute class
        func_outer = self.time_func
        def new_func(t):
            if t < start: return func_outer(t)
            elif t > end:
                if stay: return new_func(end)
                else: return func_outer(t)
            elif not lincl and t == start: return func_outer(t)
            elif not rincl and t == end: return func_outer(t)
            else: return tuple(float(i) + value[idx] for idx, i in enumerate(func_outer(t)))
        self.time_func = new_func
        self.last_change = max(self.last_change, end)

    def interpolate(self, other, start, end, easing=easings.smooth):
        """Interpolate this value from start to end with easing to the other value
        Return a lambda function (or this same attribute) for these times"""
        assert self.__class__ == Tup  # Both classes need to be a real object
        assert other.__class__ == Tup
        start_val = self.time_func(start)
        end_val = other.time_func(end)
        # We need to get from start_val to end_val with easing
        new = Tup(0, ())
        new.time_func = lambda t: tuple(s + (e-s) * easing((t-start)/(end-start)) for s, e in zip(start_val, end_val))
        new.last_change = end
        return new



class Coor(Real):
    """The coordinate attribute is a real-valued attribute (changing) through time"""
    def __init__(self, creation, start_val=(0, 0)):
        # Define the function
        self.time_func = lambda t: (
                heaviside(creation)(t) * start_val[0],
                heaviside(creation)(t) * start_val[1],
            )
        # Define the last change time of this attribute (after that it's constant)
        self.last_change = creation

    def add_onward(self, start, func, lincl=True, last_change=None):
        """Set this attribute to value + func(val) on [start, infty]
        lincl are if the bracket is [/] or (/)"""
        func_left = self.time_func
        def new_func(t):
            if t < start: return func_outer(t)
            elif not lincl and t == start: return func_left(t)
            else:
                if callable(func): return (func_left(t)[0] + func(t)[0], func_left(t)[1] + func(t)[1])
                else: return (func_left(t)[0] + func[0], func_left(t)[1] + func[1])
        self.time_func = new_func
        self.last_change = max(self.last_change, start)
        if last_change != None:
            self.last_change = max(self.last_change, last_change)

    def rotate_around(self, start_time, end_time, pivot_point, degrees, clockwise=False, stay=True):
        """Animate the object going from pivoting around the pivot point counterclockwise"""
        # NOTE: TODO: allow pivot_point to be a function
        old_func = self.time_func
        def f(t):
            now = old_func(t)
            rel_pos = (now[0] - pivot_point[0], now[1] - pivot_point[1])
            # Get the radius and angle of the point
            r = math.sqrt(rel_pos[0]**2 + rel_pos[1]**2)
            curr_angle = math.atan2(rel_pos[1], rel_pos[0]) / math.pi * 180
            # Get the current angle
            phi = lambda t: curr_angle + degrees * (t - start_time)/(end_time - start_time)
            # Get the position
            return (pivot_point[0] + r * math.cos(phi(t)/180*math.pi),
                    pivot_point[1] + r * math.sin((1 if clockwise else -1) * phi(t)/180*math.pi))
        self.set(start_time, end_time, lambda t: f(t), stay=stay)

        # if start_pos == None:
        #     now = self.time_func(start_time)
        # else:
        #     now = start_pos
        # rel_pos = (now[0] - pivot_point[0], now[1] - pivot_point[1])
        # # Get the radius and angle of the point
        # r = math.sqrt(rel_pos[0]**2 + rel_pos[1]**2)
        # curr_angle = math.atan2(rel_pos[1], rel_pos[0]) / math.pi * 180
        # # Get the current angle
        # phi = lambda t: curr_angle + degrees * (t - start_time)/(end_time - start_time)
        # # Get the position
        # self.set(start_time, end_time, lambda t: (
        #     pivot_point[0] + r * math.cos(phi(t)/180*math.pi),
        #     pivot_point[1] + r * math.sin((1 if clockwise else -1) * phi(t)/180*math.pi),
        # ), stay=stay)
        self.last_change = end_time

    def move_to(self, start_time, end_time, end_pos, stay=True, easing=easings.smooth):
        """Animate the object going from start_pos to end_pos, exponentially"""
        start_pos = self.time_func(start_time)
        dx, dy = start_pos[0]-end_pos[0], start_pos[1]-end_pos[1]

        s, e = start_time, end_time
        self.set(s, e, lambda t: (
            dx * (1-easing((t-s)/(e-s))) + end_pos[0],
            dy * (1-easing((t-s)/(e-s))) + end_pos[1],
        ), stay=stay)
        self.last_change = max(self.last_change, end_time)

    def add(self, start, end, func_inner, lincl=True, rincl=True, stay=False):
        """Set this attribute to func_inner(x)+previous on [start, end]
        lincl, rincl are if the brackets are [/] or (/)"""
        func_outer = self.time_func
        def new_func(t):
            if t < start: return func_outer(t)
            elif t > end:
                if stay: return new_func(end)
                else: return func_outer(t)
            elif not lincl and t == start: return func_outer(t)
            elif not rincl and t == end: return func_outer(t)
            else: return (float(func_outer(t)[0]) + func_inner(t)[0], float(func_outer(t)[1]) + func_inner(t)[1])
        self.time_func = new_func
        self.last_change = max(self.last_change, end_time)


class String(Real):
    """The real attribute is a real-valued attribute (changing) through time"""
    def __init__(self, creation, start_val=''):
        if isinstance(start_val, String):
            self.set_to(start_val)
        else:
            # Define the function
            self.time_func = lambda t: heaviside(creation)(t) * start_val
            # Define the last change time of this attribute (after that it's constant)
            self.last_change = creation


# :::MARKER: Optimisation is only done below


class Color:
    """Parses any input to a workable color internally
    input can be a LinearGradient, RadialGradient, rgb, rgba, name or hex value
    Can return the svg-specifyer for the color"""
    def __init__(self, creation=0, start_color='#000', use=None):
        assert isinstance(creation, int | float)
        if use != None:
            assert isinstance(use, str)
            self.use = use
        if isinstance(start_color, Color):
            self.set_to(start_color)
        else:
            self.use, col = self.parse(start_color)
            if self.use == 'rgb':
                # Define the function
                self.time_func = lambda t: (
                        heaviside(creation)(t) * col[0],
                        heaviside(creation)(t) * col[1],
                        heaviside(creation)(t) * col[2],
                    )
            elif self.use == 'rgba':
                # Define the function
                self.time_func = lambda t: (
                        heaviside(creation)(t) * col[0],
                        heaviside(creation)(t) * col[1],
                        heaviside(creation)(t) * col[2],
                        heaviside(creation)(t) * col[3],
                    )
            else:
                raise NotImplementedError('use has not implemented a time function (yet)')
        self.last_change = creation

    def set(self, start, end, func_inner, lincl=True, rincl=True, stay=False):
        """Set this attribute to func_inner(x) on [start, end]
        lincl, rincl are if the brackets are [/] or (/)"""
        # NOTE: TODO: make some assertions an the type of functions that this is
        # Do we need to parse the color? -> Yes
        # We assume for now it is rgb
        func_outer = self.time_func
        def new_func(t):
            if t < start: return func_outer(t)
            elif t > end:
                if stay: return new_func(end)
                else: return func_outer(t)
            elif not lincl and t == start: return func_outer(t)
            elif not rincl and t == end: return func_outer(t)
            else: return func_inner(t)
        self.time_func = new_func
        self.last_change = max(self.last_change, end)

    def set_to(self, other):
        """Set the value of self to always be the value of other"""
        assert isinstance(other, Color)
        self.use = other.use
        # The value needs to update as the time_function updates
        # we thus use a lambda function
        self.time_func = lambda t: other.time_func(t)
        self.last_change = other.last_change

    def parse(self, color):
        """Parses the color into usage and values"""
        if isinstance(color, colors.LinearGradient):
            raise NotImplementedError('NOTE: TODO')
        elif isinstance(color, colors.RadialGradient):
            raise NotImplementedError('NOTE: TODO')
        elif isinstance(color, tuple):
            if len(color) == 3:  # rgb value
                return ('rgb', color)
            elif len(color) == 4:  # rgba value
                return ('rgba', color)
            else:
                raise ValueError('Length of tuple does not match rgb or rgba lengths')
        elif isinstance(color, str):
            if color[0] != '#':
                # Represents a color name
                color = colors.color_from_name(color.upper())
                # Gets the hex value, thus must still be processed

            if color[0] == '#':
                # Indicator for a hex value
                if len(color) == 4:
                    # of the form #fe0
                    return ('rgb', (int(color[1]*2, 16), int(color[2]*2, 16), int(color[3]*2, 16)))
                elif len(color) == 7:
                    # of the form #ffee00
                    return ('rgb', (int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16)))
                elif len(color) == 9:
                    # of the form #ffee9900
                    return ('rgba', (int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16), int(color[7:9], 16)))
                else:
                    raise ValueError('Invalid length of hex color specifyer')
        else:
            raise ValueError('Type of color unknown')

    def random_color(self):
        # Get random value from colors
        # Like colors.random_color() or so
        raise NotImplementedError()

    def at_time(self, time, rounding_func=round):
        """Returns the svg-specifyer for this color"""
        if self.use == 'rgb':
            r, g, b = self.time_func(time)
            r, g, b = rounding_func(r), rounding_func(g), rounding_func(b)
            return f'rgb({r},{g},{b})'
        elif self.use == 'rgba':
            r, g, b, a = self.time_func(time)
            r, g, b, a = rounding_func(r), rounding_func(g), rounding_func(b), rounding_func(a)
            return f'rgba({r},{g},{b},{a})'
        else:
            raise NotImplementedError('use not implemented')

    def interpolate(self, other, start, end, easing=easings.linear):
        """Interpolate this value from start to end with easing to the other value
        Return a lambda function (or this same attribute) for these times"""
        assert self.__class__ == Color  # Both classes need to be a real object
        assert other.__class__ == Color
        if self.use == other.use == 'rgb':
            start_val = self.time_func(start)
            end_val = other.time_func(end)
            # We need to get from start_val to end_val with easing
            new = Color(use='rgb')
            new.time_func = lambda t: tuple(s + (e-s) * easing((t-start)/(end-start)) for s, e in zip(start_val, end_val))
            new.last_change = end
            return new
        else:
            raise NotImplementedError('Only rgb colors can be transformed yet.')


    def set_onward(self, start, value, lincl=True):
        """Set this attribute to value on [start, infty]
        lincl are if the bracket is [/] or (/)"""
        # NOTE: TODO: allowed value to be callable
        use, color = self.parse(value)
        assert use == self.use == 'rgb'  # NOTE: TODO: others
        func_left = self.time_func
        def new_func(t):
            if t < start: return func_left(t)
            elif not lincl and t == start: return func_left(t)
            else: return color
        self.time_func = new_func
        self.last_change = max(self.last_change, start)



import vectormation.attributes as attributes
import vectormation.easings as easings


class Styling:
    """Represents the styling part of an svg"""
    # Define the attributes that this class will have (and their respective type)
    styles = ['opacity', 'fill', 'fill_opacity', 'stroke', 'stroke_width', 'stroke_opacity',
              'fill_rule', 'stroke_dasharray', 'stroke_linecap', 'stroke_linejoin']
    transforms = ['scale_x', 'scale_y', 'dx', 'dy', 'skew_x',
                  'skew_y', 'rotation', 'skew_x_after', 'skew_y_after', 'matrix']

    def __init__(self, kwargs, creation=0, **defaults):
        # We take the shape-specific defaults and let them overwrite the global defaults
        # We take the kwargs (specified by the person animating) and let them overwrite the shape-specific defaults
        self.kw = kwargs
        self.defaults = defaults
        self.global_defaults = dict(
            opacity = 1,
            fill = '#000',
            fill_opacity = 1,
            stroke = '#000',
            stroke_width = 1,
            stroke_opacity = 1,
            fill_rule = 'nonzero',
            stroke_dasharray = '',
            stroke_linecap = 'butt',
            stroke_linejoin = 'miter',  # arcs | bevel | miter | miter-clip | round
            ### Transformations
            scale_x = 1,
            scale_y = 1,
            dx = 0,
            dy = 0,
            skew_x = 0,
            skew_y = 0,
            rotation = (0, 0, 0),
            skew_x_after = 0,
            skew_y_after = 0,
            matrix = (0, 0, 0, 0, 0, 0),
        )

        # Check for unknown specified kwargs/shape-specific defaults
        assert all(arg in self.global_defaults for arg in self.kw)  # unknown arguments passed
        assert all(arg in self.global_defaults for arg in self.defaults)
        assert all(arg in self.global_defaults for arg in self.styles + self.transforms)
        assert all((arg in self.styles) ^ (arg in self.transforms) for arg in self.global_defaults)

        # Set the values from the creation time of this Styling class onwards
        self.set_values(creation=creation, **(self.global_defaults | self.defaults | self.kw))

    def set_values(self, creation=0, **values):
        """Sets the values of the global defaults starting at the creation"""
        self.opacity            = attributes.Real(creation, values['opacity'])
        self.fill               = attributes.Color(creation, values['fill'])
        self.fill_opacity       = attributes.Real(creation, values['fill_opacity'])
        self.stroke             = attributes.Color(creation, values['stroke'])
        self.stroke_width       = attributes.Real(creation, values['stroke_width'])
        self.stroke_opacity     = attributes.Real(creation, values['stroke_opacity'])
        self.fill_rule          = attributes.String(creation, values['fill_rule'])
        self.stroke_dasharray   = attributes.String(creation, values['stroke_dasharray'])
        self.stroke_linecap     = attributes.String(creation, values['stroke_linecap'])
        self.stroke_linejoin    = attributes.String(creation, values['stroke_linejoin'])
        ### Transformations
        self.scale_x            = attributes.Real(creation, values['scale_x'])
        self.scale_y            = attributes.Real(creation, values['scale_y'])
        self.dx                 = attributes.Real(creation, values['dx'])
        self.dy                 = attributes.Real(creation, values['dy'])
        self.skew_x             = attributes.Real(creation, values['skew_x'])
        self.skew_y             = attributes.Real(creation, values['skew_y'])
        self.rotation           = attributes.Tup(creation, values['rotation'])
        self.skew_x_after       = attributes.Real(creation, values['skew_x_after'])
        self.skew_y_after       = attributes.Real(creation, values['skew_y_after'])
        self.matrix             = attributes.Tup(creation, values['matrix'])

    def kwargs(self):
        """Returns the key-word arguments specifying the styling as functions of time (attributes)"""
        return {attr: getattr(self, attr) for attr in self.global_defaults}

    @property
    def last_change(self):
        """Returns the last change of any attribute contained in this class"""
        return max(val.last_change for val in self.kwargs().values())

    def svg_style(self, time):
        """Returns the style part of the svg object"""
        # To avoid redundancy we only write each value if it is not equal to the global default value
        args = {arg: getattr(self, arg).at_time(time) for arg in self.styles}

        string = ''
        # Iterate through the pairs of names vs stylenames (the name in the svg style part)
        for name, stylename in [('opacity', 'opacity'), ('fill', 'fill'), ('fill_opacity', 'fill-opacity'),
                                ('stroke', 'stroke'), ('stroke_width', 'stroke-width'),
                                ('stroke_opacity', 'stroke-opacity'), ('fill_rule', 'fill-rule'),
                                ('stroke_dasharray', 'stroke-dasharray'), ('stroke_linecap', 'stroke-linecap'),
                                ('stroke_linejoin', 'stroke-linejoin')]:
            if args[name] != self.global_defaults[name]:
                string += f" {stylename}='{args[name]}'"

        # Now, we do the transform argument
        transform = self.transform_style(time)
        if transform != '':
            string += f" transform='{transform[1:]}'"

        return string

    def transform_style(self, time):
        """Returns the transform style argument of the styling of the svg object"""
        args = {arg: getattr(self, arg).at_time(time) for arg in self.transforms}

        transform = ''
        if not args['dx'] == args['dy'] == self.global_defaults['dx'] == self.global_defaults['dy']:
            transform += f" translate({args['dx']},{args['dy']})"
        if not args['scale_x'] == args['scale_y'] == self.global_defaults['scale_x'] == self.global_defaults['scale_y']:
            transform += f" scale({args['scale_x']},{args['scale_y']})"
        if not args['skew_x'] == self.global_defaults['skew_x']:
            transform += f" skewX({args['skew_x']})"
        if not args['skew_y'] == self.global_defaults['skew_y']:
            transform += f" skewY({args['skew_y']})"
        if not args['rotation'] == self.global_defaults['rotation']:
            r, x, y = args['rotation']
            transform += f" rotate({r%360},{x},{y})"
        if not args['skew_x_after'] == self.global_defaults['skew_x_after']:
            transform += f" skewX({args['skew_x_after']})"
        if not args['skew_y_after'] == self.global_defaults['skew_y_after']:
            transform += f" skewY({args['skew_y_after']})"
        if not args['matrix'] == self.global_defaults['matrix']:
            a, b, c, d, e, f = args['matrix']
            transform += f" matrix({a},{b},{c},{d},{e},{f})"
        return transform

    def interpolate(self, other, start, end, easing=easings.linear):
        """Interpolate all styling values from self at time start to other at time end with easing"""
        assert isinstance(other, Styling)
        # If all values are the same, i.e. the class is exactly the same, we do not need to change
        # anyting, thus return the current values.
        if self == other:
            # NOTE: TODO: also add the rotation to the fade-in objects here
            # inter_attr = self.rotation.interpolate(attributes.Tup(0, (360, 0, 0)), start, end, easing=easing)
            # self.rotation = inter_attr
            return self

        # We create a new styling class as the interpolated styling
        new_styling = Styling({}, creation=start)
        # We need to interpolate all the values of both styling classes
        # The Real attributes are easily interpolated, some others must be set at once.
        for attr in self.transforms + self.styles:
            start_attr = getattr(self, attr)
            end_attr = getattr(other, attr)
            start_val = start_attr.at_time(start)
            end_val = end_attr.at_time(end)
            if start_val == end_val:
                # if attr == 'rotation':
                #     # NOTE: TODO: add 360 degrees but rotate around the center of the object
                #     inter_attr = start_attr.interpolate(attributes.Tup(0, (360, 0, 0)), start, end, easing=easing)
                #     setattr(new_styling, attr, inter_attr)
                #     print(start_val, end_val)
                #     continue
                # We do not need to transform anything
                getattr(new_styling, attr).set(start, end, start_attr.time_func)
                continue
            # If the values are unequal we need to interpolate these values
            inter_attr = start_attr.interpolate(end_attr, start, end, easing=easing)
            setattr(new_styling, attr, inter_attr)
        # NOTE: TODO: add rotation argument that rotaetes the objects a full circle.
        return new_styling
